"""Unit tests for M365 enumerator."""

from unittest.mock import AsyncMock, MagicMock

import pytest
from msgraph_core import APIVersion

from azure_discovery.enumerators.m365 import M365Enumerator
from azure_discovery.adt_types.models import M365Config, ResourceNode


@pytest.fixture
def mock_credential():
    """Mock Azure credential."""
    return AsyncMock()


@pytest.fixture
def m365_config():
    """Create M365Config instance."""
    return M365Config(
        include_sharepoint_sites=True,
        include_teams=True,
        include_onedrive=True,
        include_exchange_mailboxes=False,
        sharepoint_max_sites=1000,
        teams_max_teams=1000,
        onedrive_max_drives=1000,
        teams_filter_archived=True,
    )


@pytest.fixture
def m365_enumerator(mock_credential, m365_config):
    """Create M365Enumerator instance."""
    return M365Enumerator(
        tenant_id="tenant-123",
        credential=mock_credential,
        config=m365_config,
    )


@pytest.fixture
def mock_sharepoint_site():
    """Create a mock SharePoint site."""
    site = MagicMock()
    site.id = "contoso.sharepoint.com,site-guid-123,web-guid-456"
    site.web_url = "https://contoso.sharepoint.com/sites/engineering"
    site.display_name = "Engineering Team Site"
    site.description = "Team site for engineering projects"
    site.created_date_time = "2024-01-15T10:30:00Z"
    site.last_modified_date_time = "2024-02-01T15:45:00Z"
    site.additional_data = {}
    return site


@pytest.fixture
def mock_team():
    """Create a mock Microsoft Team."""
    team = MagicMock()
    team.id = "team-guid-789"
    team.display_name = "Product Development"
    team.description = "Product development team collaboration"
    team.mail = "productdev@contoso.com"
    team.visibility = "Private"
    team.is_archived = False
    team.web_url = "https://teams.microsoft.com/l/team/team-guid-789"
    team.additional_data = {"resourceProvisioningOptions": ["Team"]}
    return team


@pytest.fixture
def mock_onedrive_drive():
    """Create a mock OneDrive drive."""
    drive = MagicMock()
    drive.id = "drive-guid-abc"
    drive.drive_type = "business"
    drive.name = "OneDrive - Contoso"
    drive.web_url = "https://contoso-my.sharepoint.com/personal/alice_contoso_com"
    drive.owner = MagicMock()
    drive.owner.user = MagicMock()
    drive.owner.user.id = "user-guid-alice"
    drive.owner.user.display_name = "Alice Anderson"
    drive.owner.user.email = "alice@contoso.com"
    drive.owner.user.user_principal_name = "alice@contoso.com"
    drive.owner.user.additional_data = {}
    drive.quota = MagicMock()
    drive.quota.total = 1099511627776  # 1 TB
    drive.quota.used = 549755813888    # 512 GB
    drive.quota.remaining = 549755813888
    drive.additional_data = {}
    return drive


@pytest.fixture
def mock_mailbox_user():
    """Create a mock user with mailbox."""
    user = MagicMock()
    user.id = "user-guid-bob"
    user.user_principal_name = "bob@contoso.com"
    user.mail = "bob@contoso.com"
    user.display_name = "Bob Brown"
    user.additional_data = {}
    return user


@pytest.mark.asyncio
async def test_enumerate_sharepoint_sites_success(m365_enumerator, mock_sharepoint_site):
    """Test successful enumeration of SharePoint sites."""
    mock_response = MagicMock()
    mock_response.value = [mock_sharepoint_site]
    mock_client = MagicMock()
    mock_client.sites.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    sites = await m365_enumerator.enumerate_sharepoint_sites()

    assert len(sites) == 1
    site_node = sites[0]
    assert isinstance(site_node, ResourceNode)
    assert site_node.type == "Microsoft.SharePoint/sites"
    assert site_node.name == "Engineering Team Site"
    assert site_node.properties["web_url"] == "https://contoso.sharepoint.com/sites/engineering"
    assert site_node.properties["description"] == "Team site for engineering projects"
    assert "siteId" in site_node.tags


@pytest.mark.asyncio
async def test_enumerate_sharepoint_sites_with_filter(m365_enumerator, mock_sharepoint_site):
    """Test SharePoint site enumeration with URL filter."""
    m365_enumerator.config.sharepoint_site_filter = [
        "contoso.sharepoint.com/sites/engineering"
    ]

    filtered_site = MagicMock()
    filtered_site.id = "contoso.sharepoint.com,hr-guid,web-guid"
    filtered_site.web_url = "https://contoso.sharepoint.com/sites/hr"
    filtered_site.display_name = "HR Site"
    filtered_site.name = None
    filtered_site.description = None
    filtered_site.created_date_time = None
    filtered_site.last_modified_date_time = None
    filtered_site.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_sharepoint_site, filtered_site]
    mock_client = MagicMock()
    mock_client.sites.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    sites = await m365_enumerator.enumerate_sharepoint_sites()

    assert len(sites) == 1
    assert sites[0].properties["web_url"] == "https://contoso.sharepoint.com/sites/engineering"


@pytest.mark.asyncio
async def test_enumerate_sharepoint_sites_max_limit(m365_enumerator, mock_sharepoint_site):
    """Test SharePoint site enumeration respects max limit."""
    m365_enumerator.config.sharepoint_max_sites = 2

    mock_response = MagicMock()
    mock_response.value = [mock_sharepoint_site] * 5
    mock_client = MagicMock()
    mock_client.sites.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    sites = await m365_enumerator.enumerate_sharepoint_sites()

    assert len(sites) == 2


@pytest.mark.asyncio
async def test_enumerate_teams_success(m365_enumerator, mock_team):
    """Test successful enumeration of Microsoft Teams."""
    mock_response = MagicMock()
    mock_response.value = [mock_team]
    mock_client = MagicMock()
    mock_client.groups.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    teams = await m365_enumerator.enumerate_teams()

    assert len(teams) == 1
    team_node = teams[0]
    assert isinstance(team_node, ResourceNode)
    assert team_node.type == "Microsoft.Teams/teams"
    assert team_node.name == "Product Development"
    assert team_node.properties["mail"] == "productdev@contoso.com"
    assert team_node.properties["visibility"] == "Private"
    assert team_node.properties["is_archived"] is False


@pytest.mark.asyncio
async def test_enumerate_teams_filter_archived(m365_enumerator, mock_team):
    """Test Teams enumeration filters out archived teams."""
    # Create archived team
    archived_team = MagicMock()
    archived_team.id = "archived-team-123"
    archived_team.display_name = "Archived Team"
    archived_team.is_archived = True
    archived_team.additional_data = {"resourceProvisioningOptions": ["Team"]}

    mock_response = MagicMock()
    mock_response.value = [mock_team, archived_team]
    mock_client = MagicMock()
    mock_client.groups.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    teams = await m365_enumerator.enumerate_teams()

    assert len(teams) == 1
    assert teams[0].properties["is_archived"] is False


@pytest.mark.asyncio
async def test_enumerate_teams_include_archived(m365_enumerator, mock_team):
    """Test Teams enumeration includes archived teams when filter disabled."""
    # Disable archived filter
    m365_enumerator.config.teams_filter_archived = False

    # Create archived team
    archived_team = MagicMock()
    archived_team.id = "archived-team-123"
    archived_team.display_name = "Archived Team"
    archived_team.is_archived = True
    archived_team.mail = "archived@contoso.com"
    archived_team.visibility = "Public"
    archived_team.web_url = "https://teams.microsoft.com/l/team/archived-team-123"
    archived_team.additional_data = {"resourceProvisioningOptions": ["Team"]}

    mock_response = MagicMock()
    mock_response.value = [mock_team, archived_team]
    mock_client = MagicMock()
    mock_client.groups.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    teams = await m365_enumerator.enumerate_teams()

    assert len(teams) == 2


@pytest.mark.asyncio
async def test_enumerate_onedrive_drives_success(m365_enumerator, mock_onedrive_drive):
    """Test successful enumeration of OneDrive drives."""
    mock_onedrive_drive.created_date_time = None
    mock_onedrive_drive.last_modified_date_time = None
    mock_response = MagicMock()
    mock_response.value = [mock_onedrive_drive]
    mock_client = MagicMock()
    mock_client.drives.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    drives = await m365_enumerator.enumerate_onedrive_drives()

    assert len(drives) == 1
    drive_node = drives[0]
    assert isinstance(drive_node, ResourceNode)
    assert drive_node.type == "Microsoft.OneDrive/drives"
    assert drive_node.name == "OneDrive - Contoso"
    assert drive_node.properties["drive_type"] == "business"
    assert drive_node.properties["owner_upn"] == "alice@contoso.com"
    assert drive_node.properties["quota_total"] == 1099511627776
    assert drive_node.properties["quota_used"] == 549755813888


@pytest.mark.asyncio
async def test_enumerate_onedrive_drives_with_user_filter(m365_enumerator, mock_onedrive_drive):
    """Test OneDrive enumeration with user filter."""
    # Configure user filter
    m365_enumerator.config.onedrive_user_filter = ["alice@contoso.com"]

    # Create drive that doesn't match filter
    unfiltered_drive = MagicMock()
    unfiltered_drive.id = "drive-guid-xyz"
    unfiltered_drive.name = "Bob's OneDrive"
    unfiltered_drive.owner = MagicMock()
    unfiltered_drive.owner.user = MagicMock()
    unfiltered_drive.owner.user.user_principal_name = "bob@contoso.com"
    unfiltered_drive.owner.user.additional_data = {}
    unfiltered_drive.drive_type = "business"
    unfiltered_drive.web_url = None
    unfiltered_drive.quota = None
    unfiltered_drive.created_date_time = None
    unfiltered_drive.last_modified_date_time = None

    mock_response = MagicMock()
    mock_response.value = [mock_onedrive_drive, unfiltered_drive]
    mock_client = MagicMock()
    mock_client.drives.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    drives = await m365_enumerator.enumerate_onedrive_drives()

    assert len(drives) == 1
    assert drives[0].properties["owner_upn"] == "alice@contoso.com"


@pytest.mark.asyncio
async def test_enumerate_exchange_mailboxes_success(m365_enumerator, mock_mailbox_user):
    """Test successful enumeration of Exchange mailboxes."""
    m365_enumerator.config.include_exchange_mailboxes = True
    mock_mailbox_user.display_name = "Bob Brown"

    mock_response = MagicMock()
    mock_response.value = [mock_mailbox_user]
    mock_client = MagicMock()
    mock_client.users.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    mailboxes = await m365_enumerator.enumerate_exchange_mailboxes()

    assert len(mailboxes) == 1
    mailbox_node = mailboxes[0]
    assert isinstance(mailbox_node, ResourceNode)
    assert mailbox_node.type == "Microsoft.Exchange/mailboxes"
    assert mailbox_node.name == "Bob Brown"
    assert mailbox_node.properties["user_principal_name"] == "bob@contoso.com"
    assert mailbox_node.properties["mail"] == "bob@contoso.com"


@pytest.mark.asyncio
async def test_enumerate_exchange_mailboxes_with_user_filter(m365_enumerator, mock_mailbox_user):
    """Test Exchange mailbox enumeration with user filter."""
    # Enable Exchange with user filter
    m365_enumerator.config.include_exchange_mailboxes = True
    m365_enumerator.config.exchange_user_filter = ["bob@contoso.com"]

    # Create user that doesn't match filter
    unfiltered_user = MagicMock()
    unfiltered_user.id = "user-guid-charlie"
    unfiltered_user.user_principal_name = "charlie@contoso.com"
    unfiltered_user.mail = "charlie@contoso.com"
    unfiltered_user.display_name = "Charlie Chen"
    unfiltered_user.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_mailbox_user, unfiltered_user]
    mock_client = MagicMock()
    mock_client.users.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    mailboxes = await m365_enumerator.enumerate_exchange_mailboxes()

    assert len(mailboxes) == 1
    assert mailboxes[0].properties["user_principal_name"] == "bob@contoso.com"


@pytest.mark.asyncio
async def test_enumerate_empty_response(m365_enumerator):
    """Test enumeration with empty API response."""
    mock_response = MagicMock()
    mock_response.value = []
    mock_client = MagicMock()
    mock_client.sites.get = AsyncMock(return_value=mock_response)
    m365_enumerator.graph_client = mock_client

    sites = await m365_enumerator.enumerate_sharepoint_sites()

    assert len(sites) == 0


@pytest.mark.asyncio
async def test_enumerate_sites_api_error(m365_enumerator):
    """Test SharePoint enumeration handles API errors gracefully."""
    mock_client = MagicMock()
    mock_client.sites.get = AsyncMock(side_effect=Exception("Graph API error"))
    m365_enumerator.graph_client = mock_client

    sites = await m365_enumerator.enumerate_sharepoint_sites()
    assert len(sites) == 0


@pytest.mark.asyncio
async def test_enumerate_teams_api_error(m365_enumerator):
    """Test Teams enumeration handles API errors gracefully."""
    mock_client = MagicMock()
    mock_client.groups.get = AsyncMock(side_effect=Exception("Graph API error"))
    m365_enumerator.graph_client = mock_client

    teams = await m365_enumerator.enumerate_teams()
    assert len(teams) == 0


@pytest.mark.asyncio
async def test_m365_config_defaults():
    """Test M365Config default values."""
    config = M365Config()
    
    assert config.include_sharepoint_sites is True
    assert config.include_teams is True
    assert config.include_onedrive is True
    assert config.include_exchange_mailboxes is False
    assert config.sharepoint_max_sites == 1000
    assert config.teams_max_teams == 1000
    assert config.onedrive_max_drives == 1000
    assert config.teams_filter_archived is True
    assert config.include_site_owners is True
    assert config.include_team_members is True



"""Unit tests for Purview helper functions."""

import pytest

from azure_discovery.adt_types.models import ResourceNode, ResourceRelationship
from azure_discovery.utils.purview_helpers import (
    build_purview_relationships,
    get_dlp_policies,
    get_retention_policies,
    get_sensitivity_labels,
    get_retention_labels,
    group_purview_by_type,
    filter_policies_by_location_type,
)


@pytest.fixture
def mock_sensitivity_label():
    """Create a mock sensitivity label node."""
    return ResourceNode(
        id="purview://sensitivityLabels/label-123",
        name="Confidential",
        type="Microsoft.Purview/sensitivityLabels",
        subscription_id="Tenant",
        properties={
            "description": "Confidential information",
            "color": "#FF0000",
            "sensitivity": 3,
        },
        tags={
            "labelId": "label-123",
        },
    )


@pytest.fixture
def mock_dlp_policy():
    """Create a mock DLP policy node."""
    return ResourceNode(
        id="purview://dlpPolicies/dlp-456",
        name="GDPR Compliance Policy",
        type="Microsoft.Purview/dlpPolicies",
        subscription_id="Tenant",
        properties={
            "description": "Protect GDPR sensitive data",
            "state": "enabled",
            "mode": "Enforce",
        },
        tags={
            "policyId": "dlp-456",
            "locations": [
                {"location": "https://contoso.sharepoint.com/sites/hr", "locationType": "SharePoint"},
                {"location": "All", "locationType": "Teams"},
            ],
        },
    )


@pytest.fixture
def mock_retention_policy():
    """Create a mock retention policy node."""
    return ResourceNode(
        id="purview://retentionPolicies/retention-789",
        name="7 Year Retention Policy",
        type="Microsoft.Purview/retentionPolicies",
        subscription_id="Tenant",
        properties={
            "description": "Retain records for 7 years",
            "isEnabled": True,
            "retentionDuration": 2555,
        },
        tags={
            "policyId": "retention-789",
            "locations": [
                {"location": "All", "locationType": "OneDrive"},
            ],
        },
    )


@pytest.fixture
def mock_retention_label():
    """Create a mock retention label node."""
    return ResourceNode(
        id="purview://retentionLabels/label-abc",
        name="Permanent Record",
        type="Microsoft.Purview/retentionLabels",
        subscription_id="Tenant",
        properties={
            "description": "Keep permanently",
            "isInUse": True,
        },
        tags={
            "labelId": "label-abc",
        },
    )


@pytest.fixture
def mock_sharepoint_site():
    """Create a mock SharePoint site node."""
    return ResourceNode(
        id="sharepoint://sites/site-123",
        name="HR Site",
        type="Microsoft.SharePoint/sites",
        subscription_id="Tenant",
        properties={
            "webUrl": "https://contoso.sharepoint.com/sites/hr",
        },
        tags={
            "siteId": "site-123",
        },
    )


@pytest.fixture
def mock_team():
    """Create a mock Team node."""
    return ResourceNode(
        id="teams://teams/team-456",
        name="Product Development",
        type="Microsoft.Teams/teams",
        subscription_id="Tenant",
        properties={
            "mail": "productdev@contoso.com",
        },
        tags={
            "teamId": "team-456",
        },
    )


@pytest.fixture
def mock_onedrive_drive():
    """Create a mock OneDrive drive node."""
    return ResourceNode(
        id="onedrive://drives/drive-789",
        name="Alice's OneDrive",
        type="Microsoft.OneDrive/drives",
        subscription_id="Tenant",
        properties={
            "driveType": "business",
            "ownerUpn": "alice@contoso.com",
        },
        tags={
            "driveId": "drive-789",
        },
    )


def test_get_sensitivity_labels(mock_sensitivity_label, mock_dlp_policy):
    """Test filtering nodes to sensitivity labels."""
    nodes = [mock_sensitivity_label, mock_dlp_policy]
    labels = get_sensitivity_labels(nodes)
    
    assert len(labels) == 1
    assert labels[0].type == "Microsoft.Purview/sensitivityLabels"


def test_get_dlp_policies(mock_dlp_policy, mock_retention_policy):
    """Test filtering nodes to DLP policies."""
    nodes = [mock_dlp_policy, mock_retention_policy]
    policies = get_dlp_policies(nodes)
    
    assert len(policies) == 1
    assert policies[0].type == "Microsoft.Purview/dlpPolicies"


def test_get_retention_policies(mock_dlp_policy, mock_retention_policy):
    """Test filtering nodes to retention policies."""
    nodes = [mock_dlp_policy, mock_retention_policy]
    policies = get_retention_policies(nodes)
    
    assert len(policies) == 1
    assert policies[0].type == "Microsoft.Purview/retentionPolicies"


def test_get_retention_labels(mock_retention_label, mock_sensitivity_label):
    """Test filtering nodes to retention labels."""
    nodes = [mock_retention_label, mock_sensitivity_label]
    labels = get_retention_labels(nodes)
    
    assert len(labels) == 1
    assert labels[0].type == "Microsoft.Purview/retentionLabels"


def test_group_purview_by_type(
    mock_sensitivity_label,
    mock_dlp_policy,
    mock_retention_policy,
    mock_retention_label,
):
    """Test grouping Purview nodes by type."""
    nodes = [
        mock_sensitivity_label,
        mock_dlp_policy,
        mock_retention_policy,
        mock_retention_label,
    ]
    grouped = group_purview_by_type(nodes)
    
    assert "Microsoft.Purview/sensitivityLabels" in grouped
    assert "Microsoft.Purview/dlpPolicies" in grouped
    assert "Microsoft.Purview/retentionPolicies" in grouped
    assert "Microsoft.Purview/retentionLabels" in grouped
    
    assert len(grouped["Microsoft.Purview/sensitivityLabels"]) == 1
    assert len(grouped["Microsoft.Purview/dlpPolicies"]) == 1
    assert len(grouped["Microsoft.Purview/retentionPolicies"]) == 1
    assert len(grouped["Microsoft.Purview/retentionLabels"]) == 1


def test_filter_policies_by_location_type_sharepoint(mock_dlp_policy):
    """Test filtering policies by SharePoint location type."""
    policies = [mock_dlp_policy]
    filtered = filter_policies_by_location_type(policies, "SharePoint")
    
    assert len(filtered) == 1
    assert filtered[0].id == "purview://dlpPolicies/dlp-456"


def test_filter_policies_by_location_type_teams(mock_dlp_policy):
    """Test filtering policies by Teams location type."""
    policies = [mock_dlp_policy]
    filtered = filter_policies_by_location_type(policies, "Teams")
    
    assert len(filtered) == 1
    assert filtered[0].id == "purview://dlpPolicies/dlp-456"


def test_filter_policies_by_location_type_onedrive(mock_retention_policy):
    """Test filtering policies by OneDrive location type."""
    policies = [mock_retention_policy]
    filtered = filter_policies_by_location_type(policies, "OneDrive")
    
    assert len(filtered) == 1
    assert filtered[0].id == "purview://retentionPolicies/retention-789"


def test_filter_policies_by_location_type_no_match(mock_dlp_policy):
    """Test filtering policies with no matching location type."""
    policies = [mock_dlp_policy]
    filtered = filter_policies_by_location_type(policies, "Exchange")
    
    # DLP policy has SharePoint and Teams, not Exchange
    assert len(filtered) == 0


def test_build_dlp_sharepoint_relationships(mock_dlp_policy, mock_sharepoint_site):
    """Test building DLP policy -> SharePoint site relationships."""
    purview_nodes = [mock_dlp_policy]
    m365_nodes = [mock_sharepoint_site]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    assert len(relationships) == 1
    assert relationships[0].relation_type == "applies_to"
    assert relationships[0].source_id == "purview://dlpPolicies/dlp-456"
    assert relationships[0].target_id == "sharepoint://sites/site-123"


def test_build_dlp_teams_relationships(mock_dlp_policy, mock_team):
    """Test building DLP policy -> Teams relationships."""
    purview_nodes = [mock_dlp_policy]
    m365_nodes = [mock_team]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # DLP policy has 'All' Teams, so should apply to the team
    assert len(relationships) == 1
    assert relationships[0].relation_type == "applies_to"
    assert relationships[0].source_id == "purview://dlpPolicies/dlp-456"
    assert relationships[0].target_id == "teams://teams/team-456"


def test_build_retention_onedrive_relationships(mock_retention_policy, mock_onedrive_drive):
    """Test building retention policy -> OneDrive relationships."""
    purview_nodes = [mock_retention_policy]
    m365_nodes = [mock_onedrive_drive]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # Retention policy has 'All' OneDrive, so should apply to the drive
    assert len(relationships) == 1
    assert relationships[0].relation_type == "retains"
    assert relationships[0].source_id == "purview://retentionPolicies/retention-789"
    assert relationships[0].target_id == "onedrive://drives/drive-789"


def test_build_relationships_all_location_types(
    mock_dlp_policy,
    mock_retention_policy,
    mock_sharepoint_site,
    mock_team,
    mock_onedrive_drive,
):
    """Test building relationships for all Purview policy and M365 workload types."""
    purview_nodes = [mock_dlp_policy, mock_retention_policy]
    m365_nodes = [mock_sharepoint_site, mock_team, mock_onedrive_drive]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # DLP applies to SharePoint (1) + Teams (1) = 2
    # Retention applies to OneDrive (1) = 1
    # Total = 3 relationships
    assert len(relationships) >= 3

    # Verify relationship types
    rel_types = {r.relation_type for r in relationships}
    assert "applies_to" in rel_types  # DLP policies
    assert "retains" in rel_types     # Retention policies


def test_build_relationships_specific_url_matching(mock_sharepoint_site):
    """Test DLP policy with specific URL matches only that site."""
    # Create DLP policy with specific URL (matches mock site)
    dlp_specific = ResourceNode(
        id="purview://dlpPolicies/dlp-specific",
        name="Specific DLP Policy",
        type="Microsoft.Purview/dlpPolicies",
        subscription_id="Tenant",
        properties={},
        tags={
            "policyId": "dlp-specific",
            "locations": [
                {"location": "https://contoso.sharepoint.com/sites/hr", "locationType": "SharePoint"},
            ],
        },
    )
    
    # Create site that doesn't match
    other_site = ResourceNode(
        id="sharepoint://sites/site-other",
        name="Engineering Site",
        type="Microsoft.SharePoint/sites",
        subscription_id="Tenant",
        properties={
            "webUrl": "https://contoso.sharepoint.com/sites/engineering",
        },
        tags={"siteId": "site-other"},
    )
    
    purview_nodes = [dlp_specific]
    m365_nodes = [mock_sharepoint_site, other_site]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # Should only match HR site
    assert len(relationships) == 1
    assert relationships[0].target_id == "sharepoint://sites/site-123"


def test_build_relationships_all_location_matches_all_resources(mock_team):
    """Test policy with 'All' location matches all resources of that type."""
    # Create policy with 'All' Teams location
    dlp_all_teams = ResourceNode(
        id="purview://dlpPolicies/dlp-all-teams",
        name="All Teams DLP Policy",
        type="Microsoft.Purview/dlpPolicies",
        subscription_id="Tenant",
        properties={},
        tags={
            "policyId": "dlp-all-teams",
            "locations": [
                {"location": "All", "locationType": "Teams"},
            ],
        },
    )
    
    # Create multiple teams
    team2 = ResourceNode(
        id="teams://teams/team-999",
        name="Engineering Team",
        type="Microsoft.Teams/teams",
        subscription_id="Tenant",
        properties={},
        tags={"teamId": "team-999"},
    )
    
    purview_nodes = [dlp_all_teams]
    m365_nodes = [mock_team, team2]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # Should match both teams
    assert len(relationships) == 2
    target_ids = {r.target_id for r in relationships}
    assert "teams://teams/team-456" in target_ids
    assert "teams://teams/team-999" in target_ids


def test_build_relationships_empty_purview_nodes():
    """Test building relationships with empty Purview nodes."""
    relationships, _ = build_purview_relationships(
        purview_nodes=[],
        m365_nodes=[],
    )

    assert len(relationships) == 0


def test_build_relationships_empty_m365_nodes(mock_dlp_policy):
    """Test building relationships with empty M365 nodes."""
    relationships, _ = build_purview_relationships(
        purview_nodes=[mock_dlp_policy],
        m365_nodes=[],
    )

    # No M365 resources to apply to
    assert len(relationships) == 0


def test_build_relationships_no_location_tags():
    """Test handling Purview policies without location tags."""
    # Create policy without locations tag
    policy_no_locations = ResourceNode(
        id="purview://dlpPolicies/dlp-no-locations",
        name="Policy Without Locations",
        type="Microsoft.Purview/dlpPolicies",
        subscription_id="Tenant",
        properties={},
        tags={"policyId": "dlp-no-locations"},  # No locations
    )
    
    purview_nodes = [policy_no_locations]
    m365_nodes = []
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # Should not create relationships for policies without locations
    assert len(relationships) == 0


def test_build_relationships_multiple_policies_same_resource(
    mock_dlp_policy,
    mock_retention_policy,
    mock_sharepoint_site,
):
    """Test multiple policies applying to the same resource."""
    # Create DLP policy that also applies to SharePoint
    dlp_sharepoint = ResourceNode(
        id="purview://dlpPolicies/dlp-sp",
        name="SharePoint DLP Policy",
        type="Microsoft.Purview/dlpPolicies",
        subscription_id="Tenant",
        properties={},
        tags={
            "policyId": "dlp-sp",
            "locations": [
                {"location": "All", "locationType": "SharePoint"},
            ],
        },
    )
    
    # Create retention policy that applies to SharePoint
    retention_sharepoint = ResourceNode(
        id="purview://retentionPolicies/retention-sp",
        name="SharePoint Retention Policy",
        type="Microsoft.Purview/retentionPolicies",
        subscription_id="Tenant",
        properties={},
        tags={
            "policyId": "retention-sp",
            "locations": [
                {"location": "All", "locationType": "SharePoint"},
            ],
        },
    )
    
    purview_nodes = [dlp_sharepoint, retention_sharepoint]
    m365_nodes = [mock_sharepoint_site]
    
    relationships, _ = build_purview_relationships(
        purview_nodes=purview_nodes,
        m365_nodes=m365_nodes,
    )

    # Both policies should apply to the site
    assert len(relationships) == 2
    
    # One applies_to (DLP), one retains (Retention)
    rel_types = [r.relation_type for r in relationships]
    assert "applies_to" in rel_types
    assert "retains" in rel_types


def test_filtering_helpers_empty_input():
    """Test filtering helpers with empty input."""
    assert len(get_sensitivity_labels([])) == 0
    assert len(get_dlp_policies([])) == 0
    assert len(get_retention_policies([])) == 0
    assert len(get_retention_labels([])) == 0


def test_group_purview_by_type_empty_input():
    """Test grouping with empty input."""
    grouped = group_purview_by_type([])
    assert len(grouped) == 0


def test_filter_policies_by_location_type_empty_input():
    """Test location type filtering with empty input."""
    filtered = filter_policies_by_location_type([], "SharePoint")
    assert len(filtered) == 0



"""M365Enumerator for Microsoft 365 workload discovery."""

import logging
from typing import Any, Dict, List, Optional

from azure.identity import DefaultAzureCredential
from msgraph import GraphServiceClient
from msgraph.generated.models.o_data_errors.o_data_error import ODataError

from azure_discovery.adt_types.models import M365Config, ResourceNode

logger = logging.getLogger(__name__)


class M365Enumerator:
    """Enumerates Microsoft 365 resources (SharePoint, Teams, OneDrive, Exchange)."""

    def __init__(
        self,
        tenant_id: str,
        credential: Optional[DefaultAzureCredential] = None,
        config: Optional[M365Config] = None,
    ):
        """Initialize M365 enumerator.

        Args:
            tenant_id: Azure AD tenant ID
            credential: Azure credential for authentication (uses DefaultAzureCredential if not provided)
            config: M365 enumeration configuration
        """
        self.tenant_id = tenant_id
        self.credential = credential or DefaultAzureCredential()
        self.config = config or M365Config()

        # Initialize Microsoft Graph client
        self.graph_client = GraphServiceClient(self.credential, scopes=["https://graph.microsoft.com/.default"])

        logger.info(
            f"Initialized M365Enumerator for tenant {tenant_id} "
            f"(SharePoint: {self.config.include_sharepoint_sites}, "
            f"Teams: {self.config.include_teams}, "
            f"OneDrive: {self.config.include_onedrive}, "
            f"Exchange: {self.config.include_exchange_mailboxes})"
        )

    async def enumerate_sharepoint_sites(self) -> List[ResourceNode]:
        """Enumerate SharePoint sites.

        Returns:
            List of ResourceNode objects representing SharePoint sites
        """
        if not self.config.include_sharepoint_sites:
            logger.debug("SharePoint site enumeration disabled")
            return []

        logger.info("Enumerating SharePoint sites...")
        sites: List[ResourceNode] = []
        site_count = 0

        try:
            # Get all sites (paginated)
            result = await self.graph_client.sites.get()

            if not result or not result.value:
                logger.warning("No SharePoint sites found")
                return []

            for site in result.value:
                if site_count >= self.config.sharepoint_max_sites:
                    logger.info(
                        f"Reached max SharePoint sites limit ({self.config.sharepoint_max_sites}), stopping enumeration"
                    )
                    break

                # Apply site filter if configured
                if self.config.sharepoint_site_filter:
                    if not any(pattern in (site.web_url or "") for pattern in self.config.sharepoint_site_filter):
                        continue

                # Build ResourceNode
                node = ResourceNode(
                    id=f"https://graph.microsoft.com/v1.0/sites/{site.id}",
                    name=site.display_name or site.name or "Unknown Site",
                    type="Microsoft.SharePoint/sites",
                    subscription_id=self.tenant_id,  # M365 resources use tenant_id as subscription_id
                    location="global",
                    properties={
                        "site_id": site.id,
                        "web_url": site.web_url,
                        "description": site.description,
                        "created_datetime": str(site.created_date_time) if site.created_date_time else None,
                        "last_modified_datetime": str(site.last_modified_date_time)
                        if site.last_modified_date_time
                        else None,
                    },
                    tags={"siteId": site.id},
                )

                sites.append(node)
                site_count += 1

            logger.info(f"Enumerated {len(sites)} SharePoint sites")

        except ODataError as e:
            logger.error(f"Error enumerating SharePoint sites: {e.error.message if e.error else str(e)}")
        except Exception as e:
            logger.error(f"Unexpected error enumerating SharePoint sites: {e}")

        return sites

    async def enumerate_teams(self) -> List[ResourceNode]:
        """Enumerate Microsoft Teams.

        Returns:
            List of ResourceNode objects representing Teams
        """
        if not self.config.include_teams:
            logger.debug("Teams enumeration disabled")
            return []

        logger.info("Enumerating Microsoft Teams...")
        teams: List[ResourceNode] = []
        team_count = 0

        try:
            # Get all teams (groups with resourceProvisioningOptions containing 'Team')
            result = await self.graph_client.groups.get(
                request_configuration=lambda req_config: (
                    req_config.query_parameters.update(
                        {
                            "$filter": "resourceProvisioningOptions/any(x:x eq 'Team')",
                            "$select": "id,displayName,description,mail,visibility,createdDateTime,isArchived,resourceProvisioningOptions",
                        }
                    )
                )
            )

            if not result or not result.value:
                logger.warning("No Teams found")
                return []

            for team in result.value:
                if team_count >= self.config.teams_max_teams:
                    logger.info(f"Reached max Teams limit ({self.config.teams_max_teams}), stopping enumeration")
                    break

                # Filter archived teams if configured
                if self.config.teams_filter_archived and team.is_archived:
                    continue

                # Build ResourceNode
                node = ResourceNode(
                    id=f"https://graph.microsoft.com/v1.0/teams/{team.id}",
                    name=team.display_name or "Unknown Team",
                    type="Microsoft.Teams/teams",
                    subscription_id=self.tenant_id,
                    location="global",
                    properties={
                        "team_id": team.id,
                        "description": team.description,
                        "mail": team.mail,
                        "visibility": team.visibility,
                        "created_datetime": str(team.created_date_time) if team.created_date_time else None,
                        "is_archived": team.is_archived,
                    },
                )

                teams.append(node)
                team_count += 1

            logger.info(f"Enumerated {len(teams)} Microsoft Teams")

        except ODataError as e:
            logger.error(f"Error enumerating Teams: {e.error.message if e.error else str(e)}")
        except Exception as e:
            logger.error(f"Unexpected error enumerating Teams: {e}")

        return teams

    async def enumerate_onedrive_drives(self) -> List[ResourceNode]:
        """Enumerate OneDrive drives.

        Returns:
            List of ResourceNode objects representing OneDrive drives
        """
        if not self.config.include_onedrive:
            logger.debug("OneDrive enumeration disabled")
            return []

        logger.info("Enumerating OneDrive drives...")
        drives: List[ResourceNode] = []
        drive_count = 0

        try:
            # Get all drives (personal OneDrive drives)
            result = await self.graph_client.drives.get()

            if not result or not result.value:
                logger.warning("No OneDrive drives found")
                return []

            for drive in result.value:
                if drive_count >= self.config.onedrive_max_drives:
                    logger.info(
                        f"Reached max OneDrive drives limit ({self.config.onedrive_max_drives}), stopping enumeration"
                    )
                    break

                # Apply user filter if configured
                if self.config.onedrive_user_filter and drive.owner:
                    owner_upn = getattr(drive.owner.user, "user_principal_name", None)
                    if owner_upn not in self.config.onedrive_user_filter:
                        continue

                # Build ResourceNode
                node = ResourceNode(
                    id=f"https://graph.microsoft.com/v1.0/drives/{drive.id}",
                    name=drive.name or drive.id or "Unknown Drive",
                    type="Microsoft.OneDrive/drives",
                    subscription_id=self.tenant_id,
                    location="global",
                    properties={
                        "drive_id": drive.id,
                        "drive_type": drive.drive_type,
                        "web_url": drive.web_url,
                        "quota_total": drive.quota.total if drive.quota else None,
                        "quota_used": drive.quota.used if drive.quota else None,
                        "quota_remaining": drive.quota.remaining if drive.quota else None,
                        "created_datetime": str(drive.created_date_time) if drive.created_date_time else None,
                        "last_modified_datetime": str(drive.last_modified_date_time)
                        if drive.last_modified_date_time
                        else None,
                        "owner_upn": getattr(drive.owner.user, "user_principal_name", None) if drive.owner else None,
                    },
                )

                drives.append(node)
                drive_count += 1

            logger.info(f"Enumerated {len(drives)} OneDrive drives")

        except ODataError as e:
            logger.error(f"Error enumerating OneDrive drives: {e.error.message if e.error else str(e)}")
        except Exception as e:
            logger.error(f"Unexpected error enumerating OneDrive drives: {e}")

        return drives

    async def enumerate_exchange_mailboxes(self) -> List[ResourceNode]:
        """Enumerate Exchange Online mailboxes.

        Note: Requires Mail.Read.All or Exchange.ManageAsApp permission.
        This is often not granted for discovery scenarios.

        Returns:
            List of ResourceNode objects representing Exchange mailboxes
        """
        if not self.config.include_exchange_mailboxes:
            logger.debug("Exchange mailbox enumeration disabled")
            return []

        logger.warning(
            "Exchange mailbox enumeration requires Mail.Read.All or Exchange.ManageAsApp permission. "
            "This is typically not granted for discovery scenarios."
        )

        logger.info("Enumerating Exchange mailboxes...")
        mailboxes: List[ResourceNode] = []
        mailbox_count = 0

        try:
            # Get all users with mailboxes
            result = await self.graph_client.users.get(
                request_configuration=lambda req_config: (
                    req_config.query_parameters.update(
                        {
                            "$filter": "assignedLicenses/$count ne 0 and accountEnabled eq true",
                            "$select": "id,userPrincipalName,displayName,mail,mailboxSettings",
                            "$count": "true",
                        }
                    )
                )
            )

            if not result or not result.value:
                logger.warning("No Exchange mailboxes found")
                return []

            for user in result.value:
                if mailbox_count >= self.config.exchange_max_mailboxes:
                    logger.info(
                        f"Reached max Exchange mailboxes limit ({self.config.exchange_max_mailboxes}), stopping enumeration"
                    )
                    break

                # Apply user filter if configured
                if self.config.exchange_user_filter and user.user_principal_name not in self.config.exchange_user_filter:
                    continue

                # Skip users without mail
                if not user.mail:
                    continue

                # Build ResourceNode
                node = ResourceNode(
                    id=f"https://graph.microsoft.com/v1.0/users/{user.id}/mailbox",
                    name=user.display_name or user.user_principal_name or "Unknown Mailbox",
                    type="Microsoft.Exchange/mailboxes",
                    subscription_id=self.tenant_id,
                    location="global",
                    properties={
                        "user_id": user.id,
                        "user_principal_name": user.user_principal_name,
                        "mail": user.mail,
                        "mailbox_settings": user.mailbox_settings,
                    },
                )

                mailboxes.append(node)
                mailbox_count += 1

            logger.info(f"Enumerated {len(mailboxes)} Exchange mailboxes")

        except ODataError as e:
            if "Authorization_RequestDenied" in str(e) or "Insufficient privileges" in str(e):
                logger.warning(
                    f"Insufficient permissions to enumerate Exchange mailboxes: {e.error.message if e.error else str(e)}"
                )
            else:
                logger.error(f"Error enumerating Exchange mailboxes: {e.error.message if e.error else str(e)}")
        except Exception as e:
            logger.error(f"Unexpected error enumerating Exchange mailboxes: {e}")

        return mailboxes



"""Unit tests for authentication providers."""

import pytest
from unittest.mock import AsyncMock, MagicMock, patch
from azure_discovery.auth import AzureADAuthProvider, APIKeyAuthProvider
from azure_discovery.auth.models import AuthenticationError


class TestAPIKeyAuthProvider:
    """Tests for API key authentication."""
    
    def test_hash_key(self):
        """Test key hashing is consistent."""
        key = "test-key-12345"
        hash1 = APIKeyAuthProvider._hash_key(key)
        hash2 = APIKeyAuthProvider._hash_key(key)
        
        assert hash1 == hash2
        assert len(hash1) == 64  # SHA-256 hex digest
    
    def test_generate_key(self):
        """Test key generation produces unique keys."""
        key1 = APIKeyAuthProvider.generate_key()
        key2 = APIKeyAuthProvider.generate_key()
        
        assert key1 != key2
        assert len(key1) > 32
    
    @pytest.mark.asyncio
    async def test_authenticate_valid_key(self):
        """Test successful authentication with valid key."""
        key = "test-key-12345"
        key_hash = APIKeyAuthProvider._hash_key(key)
        
        provider = APIKeyAuthProvider(
            keys={
                key_hash: {
                    "id": "test-001",
                    "name": "Test Key",
                    "scopes": ["Discovery.Read"],
                    "roles": ["Reader"],
                }
            }
        )
        
        result = await provider.authenticate(key)
        
        assert result["id"] == "test-001"
        assert result["name"] == "Test Key"
        assert result["scopes"] == ["Discovery.Read"]
        assert result["source"] == "api_key"
    
    @pytest.mark.asyncio
    async def test_authenticate_invalid_key(self):
        """Test authentication fails with invalid key."""
        provider = APIKeyAuthProvider(keys={})
        
        with pytest.raises(AuthenticationError, match="Invalid API key"):
            await provider.authenticate("invalid-key")
    
    @pytest.mark.asyncio
    async def test_authenticate_revoked_key(self):
        """Test authentication fails with revoked key."""
        key = "revoked-key"
        key_hash = APIKeyAuthProvider._hash_key(key)
        
        provider = APIKeyAuthProvider(
            keys={
                key_hash: {
                    "id": "revoked-001",
                    "name": "Revoked Key",
                    "revoked": True,
                }
            }
        )
        
        with pytest.raises(AuthenticationError, match="revoked"):
            await provider.authenticate(key)


class TestAzureADAuthProvider:
    """Tests for Azure AD authentication."""
    
    @pytest.mark.asyncio
    async def test_authenticate_valid_token(self):
        """Test successful authentication with valid JWT."""
        provider = AzureADAuthProvider(
            tenant_id="test-tenant",
            client_id="test-client",
            audience="api://test",
        )
        
        # Mock the JWT validation
        mock_payload = {
            "oid": "user-123",
            "upn": "user@example.com",
            "name": "Test User",
            "roles": ["Discovery.Reader"],
            "scp": "Discovery.Read",
            "tid": "test-tenant",
        }
        
        with patch('jwt.decode', return_value=mock_payload):
            with patch.object(provider, 'jwk_client') as mock_jwk:
                mock_key = MagicMock()
                mock_key.key = "test-key"
                mock_jwk.get_signing_key_from_jwt.return_value = mock_key
                
                result = await provider.authenticate("fake-jwt-token")
        
        assert result["id"] == "user-123"
        assert result["email"] == "user@example.com"
        assert result["name"] == "Test User"
        assert result["roles"] == ["Discovery.Reader"]
        assert result["scopes"] == ["Discovery.Read"]
        assert result["source"] == "azure_ad"
    
    @pytest.mark.asyncio
    async def test_authenticate_expired_token(self):
        """Test authentication fails with expired token."""
        import jwt as jwt_module
        
        provider = AzureADAuthProvider(
            tenant_id="test-tenant",
            client_id="test-client",
            audience="api://test",
        )
        
        with patch('jwt.decode', side_effect=jwt_module.ExpiredSignatureError()):
            with patch.object(provider, 'jwk_client') as mock_jwk:
                mock_key = MagicMock()
                mock_key.key = "test-key"
                mock_jwk.get_signing_key_from_jwt.return_value = mock_key
                
                with pytest.raises(AuthenticationError, match="expired"):
                    await provider.authenticate("expired-token")
    
    @pytest.mark.asyncio
    async def test_authenticate_wrong_audience(self):
        """Test authentication fails with wrong audience."""
        import jwt as jwt_module
        
        provider = AzureADAuthProvider(
            tenant_id="test-tenant",
            client_id="test-client",
            audience="api://test",
        )
        
        with patch('jwt.decode', side_effect=jwt_module.InvalidAudienceError()):
            with patch.object(provider, 'jwk_client') as mock_jwk:
                mock_key = MagicMock()
                mock_key.key = "test-key"
                mock_jwk.get_signing_key_from_jwt.return_value = mock_key
                
                with pytest.raises(AuthenticationError, match="Invalid audience"):
                    await provider.authenticate("wrong-audience-token")



"""Helpers for reading Azure Discovery configuration from files.

Configuration files map directly onto the AzureDiscoveryRequest model.
Supported formats: JSON (.json), TOML (.toml), YAML (.yaml/.yml).
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict

from ..adt_types import AzureDiscoveryRequest


def deep_merge(base: Dict[str, Any], override: Dict[str, Any]) -> Dict[str, Any]:
    """Recursively merge dictionaries (override wins)."""

    merged: Dict[str, Any] = dict(base)
    for key, value in override.items():
        if isinstance(value, dict) and isinstance(merged.get(key), dict):
            merged[key] = deep_merge(merged[key], value)
        else:
            merged[key] = value
    return merged


def _load_json(text: str) -> Dict[str, Any]:
    payload = json.loads(text) if text.strip() else {}
    if payload is None:
        return {}
    if not isinstance(payload, dict):
        raise ValueError("JSON config must be an object at the top level")
    return payload


def _load_toml(text: str) -> Dict[str, Any]:
    import tomllib

    payload = tomllib.loads(text) if text.strip() else {}
    if payload is None:
        return {}
    if not isinstance(payload, dict):
        raise ValueError("TOML config must be a table at the top level")
    return payload


def _load_yaml(text: str) -> Dict[str, Any]:
    try:
        import yaml
    except ImportError as exc:  # pragma: no cover
        raise RuntimeError(
            "YAML config support requires PyYAML (pip install azure-discovery or pip install pyyaml)."
        ) from exc

    payload = yaml.safe_load(text) if text.strip() else {}
    if payload is None:
        return {}
    if not isinstance(payload, dict):
        raise ValueError("YAML config must be a mapping/object at the top level")
    return payload


def load_config_file(path: Path) -> Dict[str, Any]:
    """Load JSON/TOML/YAML config into a dict."""

    if not path:
        raise ValueError("path is required")

    resolved = Path(path).expanduser()
    text = resolved.read_text(encoding="utf-8")
    suffix = resolved.suffix.lower()

    if suffix == ".json":
        return _load_json(text)
    if suffix == ".toml":
        return _load_toml(text)
    if suffix in {".yaml", ".yml"}:
        return _load_yaml(text)

    # Fallback: try formats in a reasonable order.
    last_error: Exception | None = None
    for loader in (_load_json, _load_toml, _load_yaml):
        try:
            return loader(text)
        except Exception as exc:  # noqa: BLE001
            last_error = exc

    raise ValueError(f"Unsupported config format for {path}") from last_error


def load_request_from_file(path: Path) -> AzureDiscoveryRequest:
    """Load an AzureDiscoveryRequest from a config file."""

    payload = load_config_file(path)
    return AzureDiscoveryRequest.model_validate(payload)


def write_config_file(path: Path, data: Dict[str, Any], format: str = "json") -> None:
    """Write a config dict to JSON or YAML.

    Args:
        path: Output file path (parent directories created if needed).
        data: Config dict (e.g. AzureDiscoveryRequest.model_dump(mode='json')).
        format: 'json' or 'yaml'.

    Raises:
        ValueError: If format is not json or yaml.
    """
    path = Path(path).expanduser()
    path.parent.mkdir(parents=True, exist_ok=True)

    if format.lower() == "json":
        text = json.dumps(data, indent=2)
    elif format.lower() in ("yaml", "yml"):
        try:
            import yaml
        except ImportError as exc:  # pragma: no cover
            raise RuntimeError(
                "YAML config support requires PyYAML (pip install pyyaml)."
            ) from exc
        text = yaml.safe_dump(data, default_flow_style=False, sort_keys=False)
    else:
        raise ValueError(f"Unsupported write format: {format}. Use 'json' or 'yaml'.")

    path.write_text(text, encoding="utf-8")



from __future__ import annotations

from typing import Any

import pytest

from azure_discovery.adt_types import AzureDiscoveryRequest, ResourceNode
from azure_discovery.enumerators import policy_compliance


def test_extract_controls_from_compliance_metadata_dicts() -> None:
    metadata: dict[str, Any] = {
        "compliance": {
            "complianceType": "NIST SP 800-171 Rev. 2",
            "controls": [
                {
                    "controlId": "3.1.1",
                    "controlName": "Limit system access",
                    "description": "Limit system access to authorized users.",
                }
            ],
        }
    }

    controls = policy_compliance._extract_compliance_controls(metadata)
    assert len(controls) == 1
    assert controls[0].standard == "NIST SP 800-171 Rev. 2"
    assert controls[0].control_id == "3.1.1"
    assert controls[0].control_name == "Limit system access"
    assert controls[0].description == "Limit system access to authorized users."


def test_build_mapping_expands_initiative_member_policies_with_reference_id() -> None:
    initiative_id = "/providers/Microsoft.Authorization/policySetDefinitions/nist800171r2"
    policy_id = "/providers/Microsoft.Authorization/policyDefinitions/policy1"

    definitions: dict[str, dict[str, Any]] = {
        initiative_id.lower(): {
            "name": "nist800171r2",
            "type": "microsoft.authorization/policysetdefinitions",
            "properties": {
                "displayName": "NIST SP 800-171 Rev. 2",
                "description": "NIST 800-171 R2 initiative",
                "metadata": {"category": "Regulatory Compliance"},
                "policyDefinitions": [
                    {
                        "policyDefinitionId": policy_id,
                        "policyDefinitionReferenceId": "ref-3.1.1",
                    }
                ],
            },
        },
        policy_id.lower(): {
            "name": "policy1",
            "type": "microsoft.authorization/policydefinitions",
            "properties": {
                "displayName": "Example policy",
                "metadata": {
                    "compliance": {
                        "complianceType": "NIST SP 800-171 Rev. 2",
                        "controls": ["3.1.1"],
                    }
                },
            },
        },
    }

    assignment: dict[str, Any] = {
        "id": "/subscriptions/0000/providers/Microsoft.Authorization/policyAssignments/a1",
        "name": "a1",
        "properties": {
            "policyDefinitionId": initiative_id,
            "displayName": "Assign NIST 800-171",
            "scope": "/subscriptions/0000",
        },
    }

    mapping = policy_compliance._build_compliance_mapping(assignment, definitions)
    assert mapping is not None
    assert mapping.initiative_id == initiative_id.lower()
    assert mapping.initiative_name == "NIST SP 800-171 Rev. 2"

    assert any(c.control_id == "3.1.1" for c in mapping.controls)
    assert any(
        e.control_id == "3.1.1"
        and e.policy_definition_id == policy_id.lower()
        and e.policy_definition_reference_id == "ref-3.1.1"
        for e in mapping.control_evidence
    )


@pytest.mark.asyncio
async def test_enrich_sets_flattened_noncompliance_reference_ids(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def fake_assignments(*_: Any, **__: Any) -> list[dict[str, Any]]:
        return [
            {
                "id": "/subscriptions/0000/providers/Microsoft.Authorization/policyAssignments/a1",
                "name": "a1",
                "scope": "/subscriptions/0000",
                "properties": {
                    "policyDefinitionId": "/providers/Microsoft.Authorization/policyDefinitions/policy1",
                    "displayName": "Assign policy1",
                    "scope": "/subscriptions/0000",
                },
            }
        ]

    async def fake_definitions(*_: Any, **__: Any) -> dict[str, dict[str, Any]]:
        policy_id = "/providers/microsoft.authorization/policydefinitions/policy1"
        return {
            policy_id: {
                "name": "policy1",
                "type": "microsoft.authorization/policydefinitions",
                "properties": {
                    "displayName": "Example policy",
                    "metadata": {
                        "compliance": {
                            "complianceType": "NIST SP 800-171 Rev. 2",
                            "controls": ["3.1.1"],
                        }
                    },
                },
            }
        }

    async def fake_states(*_: Any, **__: Any) -> list[dict[str, Any]]:
        return [
            {
                "resourceId": "/subscriptions/0000/resourcegroups/rg/providers/microsoft.storage/storageaccounts/sa1",
                "policyAssignmentId": "/subscriptions/0000/providers/microsoft.authorization/policyassignments/a1",
                "policyDefinitionId": "/providers/microsoft.authorization/policydefinitions/policy1",
                "policyDefinitionReferenceId": "ref-a",
                "complianceState": "NonCompliant",
            },
            {
                "resourceId": "/subscriptions/0000/resourcegroups/rg/providers/microsoft.storage/storageaccounts/sa1",
                "policyAssignmentId": "/subscriptions/0000/providers/microsoft.authorization/policyassignments/a1",
                "policyDefinitionId": "/providers/microsoft.authorization/policydefinitions/policy1",
                "policyDefinitionReferenceId": "ref-b",
                "complianceState": "NonCompliant",
            },
        ]

    monkeypatch.setattr(policy_compliance, "_query_policy_assignments", fake_assignments)
    monkeypatch.setattr(policy_compliance, "_query_policy_definitions", fake_definitions)
    monkeypatch.setattr(policy_compliance, "_query_policy_states", fake_states)

    nodes = [
        ResourceNode(
            id="/subscriptions/0000/resourceGroups/rg/providers/Microsoft.Storage/storageAccounts/sa1",
            name="sa1",
            type="Microsoft.Storage/storageAccounts",
            subscription_id="0000",
            resource_group="rg",
        )
    ]

    request = AzureDiscoveryRequest(tenant_id="t")

    enriched = await policy_compliance.enrich_resources_with_compliance(
        nodes=nodes,
        request=request,
        credential=object(),
        subscriptions=["0000"],
        base_url="https://management.azure.com/",
    )

    assert len(enriched) == 1
    assert len(enriched[0].compliance_mappings) == 1

    mapping = enriched[0].compliance_mappings[0]
    assert mapping.compliance_state == "NonCompliant"
    assert len(mapping.noncompliance_evidence) == 2

    assert mapping.noncompliant_reference_ids == ["ref-a", "ref-b"]
    assert mapping.noncompliant_policy_definition_ids == [
        "/providers/microsoft.authorization/policydefinitions/policy1"
    ]


@pytest.mark.asyncio
async def test_list_policy_initiatives_uses_resource_graph(monkeypatch: pytest.MonkeyPatch) -> None:
    async def fake_query_resource_graph(*_: Any, **__: Any) -> list[dict[str, Any]]:
        return [
            {
                "id": "/providers/Microsoft.Authorization/policySetDefinitions/nist800171r2",
                "name": "nist800171r2",
                "displayName": "NIST SP 800-171 Rev. 2",
                "category": "Regulatory Compliance",
                "policyType": "BuiltIn",
            }
        ]

    monkeypatch.setattr(policy_compliance, "query_resource_graph", fake_query_resource_graph)

    initiatives = await policy_compliance.list_policy_initiatives(
        credential=object(),
        subscriptions=["0000"],
        base_url="https://management.azure.com/",
        batch_size=1000,
    )

    assert len(initiatives) == 1
    assert initiatives[0].display_name == "NIST SP 800-171 Rev. 2"



"""Integration-style tests for the discovery orchestrator."""

from __future__ import annotations

import pytest

from azure_discovery.adt_types import (
    AzureDiscoveryRequest,
    AzureDiscoveryResponse,
    DiscoveryMergeMetrics,
    ResourceNode,
    ResourceRelationship,
    VisualizationOptions,
    VisualizationResponse,
)
from azure_discovery.orchestrator import run_discovery


@pytest.mark.asyncio
async def test_run_discovery_enriches_response(monkeypatch: pytest.MonkeyPatch) -> None:
    request = AzureDiscoveryRequest(tenant_id="tenant-1")
    base_response = AzureDiscoveryResponse(
        tenant_id="tenant-1",
        discovered_subscriptions=["sub-a"],
        nodes=[
            ResourceNode(
                id="vm-a",
                name="vm-a",
                type="Microsoft.Compute/virtualMachines",
                subscription_id="sub-a",
            )
        ],
        relationships=[],
        total_resources=1,
    )

    monkeypatch.setattr(
        "azure_discovery.orchestrator.build_environment_config", lambda env: object()
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.get_credential", lambda req, cfg: object()
    )

    async def fake_enumerate(
        req: AzureDiscoveryRequest, credential=None
    ) -> AzureDiscoveryResponse:
        assert req is request
        return base_response

    def fake_render_visualization(
        response: AzureDiscoveryResponse,
        options: VisualizationOptions,
    ) -> VisualizationResponse:
        assert response is base_response
        assert options == request.visualization
        return VisualizationResponse(
            html_path="artifacts/graphs/test.html",
            nodes=len(response.nodes),
            relationships=len(response.relationships),
        )

    emitted: dict[str, AzureDiscoveryResponse] = {}

    def fake_emit_console_summary(response: AzureDiscoveryResponse) -> None:
        emitted["response"] = response

    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_azure_resources", fake_enumerate
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.render_visualization", fake_render_visualization
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.emit_console_summary", fake_emit_console_summary
    )

    enriched_response = await run_discovery(request)

    assert enriched_response.html_report_path == "artifacts/graphs/test.html"
    assert enriched_response is not base_response
    assert emitted["response"] is enriched_response


@pytest.mark.asyncio
async def test_run_discovery_includes_metrics_when_enabled(monkeypatch: pytest.MonkeyPatch) -> None:
    request = AzureDiscoveryRequest(tenant_id="tenant-1", include_entra=True, include_metrics=True)

    base_response = AzureDiscoveryResponse(
        tenant_id="tenant-1",
        discovered_subscriptions=["sub-a"],
        nodes=[],
        relationships=[],
        total_resources=0,
    )

    entra_nodes = [
        ResourceNode(
            id="graph://user/u1",
            name="User",
            type="Microsoft.Graph/User",
            subscription_id="Tenant",
            tags={"graph_id": "u1"},
        )
    ]
    entra_relationships = [
        ResourceRelationship(
            source_id="graph://user/u1",
            target_id="graph://user/u1",
            relation_type="self",
            weight=1.0,
        )
    ]

    monkeypatch.setattr(
        "azure_discovery.orchestrator.build_environment_config", lambda env: object()
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.get_credential", lambda req, cfg: object()
    )

    async def fake_enumerate_azure(req: AzureDiscoveryRequest, credential=None) -> AzureDiscoveryResponse:
        return base_response

    async def fake_enumerate_entra(req: AzureDiscoveryRequest, credential=None):
        return entra_nodes, entra_relationships

    def fake_render_visualization(response: AzureDiscoveryResponse, options: VisualizationOptions):
        return VisualizationResponse(html_path="x.html", nodes=0, relationships=0)

    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_azure_resources", fake_enumerate_azure
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_entra_resources", fake_enumerate_entra
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.render_visualization", fake_render_visualization
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.emit_console_summary", lambda resp: None
    )

    out = await run_discovery(request)

    assert out.metrics is not None
    assert out.metrics.merge == DiscoveryMergeMetrics(
        dropped_dangling_relationships=0,
        deduped_relationships=0,
    )
    assert [p.name for p in out.metrics.phases] == [
        "azure.arm",
        "entra.graph",
        "report.render_html",
    ]


@pytest.mark.asyncio
async def test_run_discovery_materializes_missing_endpoints(monkeypatch: pytest.MonkeyPatch) -> None:
    request = AzureDiscoveryRequest(
        tenant_id="tenant-1",
        include_entra=True,
        include_metrics=True,
        materialize_missing_endpoints=True,
    )

    base_response = AzureDiscoveryResponse(
        tenant_id="tenant-1",
        discovered_subscriptions=["sub-a"],
        nodes=[],
        relationships=[],
        total_resources=0,
    )

    # Relationship references endpoints that aren't present as nodes.
    entra_nodes = []
    entra_relationships = [
        ResourceRelationship(
            source_id="graph://user/u1",
            target_id="graph://user/u2",
            relation_type="member",
            weight=1.0,
        )
    ]

    monkeypatch.setattr(
        "azure_discovery.orchestrator.build_environment_config", lambda env: object()
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.get_credential", lambda req, cfg: object()
    )

    async def fake_enumerate_azure(req: AzureDiscoveryRequest, credential=None) -> AzureDiscoveryResponse:
        return base_response

    async def fake_enumerate_entra(req: AzureDiscoveryRequest, credential=None):
        return entra_nodes, entra_relationships

    def fake_render_visualization(response: AzureDiscoveryResponse, options: VisualizationOptions):
        return VisualizationResponse(html_path="x.html", nodes=0, relationships=0)

    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_azure_resources", fake_enumerate_azure
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_entra_resources", fake_enumerate_entra
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.render_visualization", fake_render_visualization
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.emit_console_summary", lambda resp: None
    )

    out = await run_discovery(request)

    assert out.metrics is not None
    assert out.metrics.merge.dropped_dangling_relationships == 0
    assert len(out.relationships) == 1
    assert len(out.nodes) == 2


@pytest.mark.asyncio
async def test_run_discovery_defender_phase(monkeypatch: pytest.MonkeyPatch) -> None:
    """Run discovery with include_defender_cloud to cover Defender phase and merge."""
    request = AzureDiscoveryRequest(
        tenant_id="tenant-1",
        include_defender_cloud=True,
        include_metrics=True,
    )
    base_response = AzureDiscoveryResponse(
        tenant_id="tenant-1",
        discovered_subscriptions=["sub-a"],
        nodes=[
            ResourceNode(
                id="/subscriptions/sub-a/resourceGroups/rg/providers/Microsoft.Compute/virtualMachines/vm1",
                name="vm1",
                type="Microsoft.Compute/virtualMachines",
                subscription_id="sub-a",
            )
        ],
        relationships=[],
        total_resources=1,
    )

    mock_config = type("Config", (), {"resource_manager": "https://management.azure.com/"})()

    monkeypatch.setattr(
        "azure_discovery.orchestrator.build_environment_config",
        lambda env: mock_config,
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.get_credential",
        lambda req, cfg: object(),
    )

    async def fake_enumerate_azure(
        req: AzureDiscoveryRequest, credential=None
    ) -> AzureDiscoveryResponse:
        return base_response

    def fake_render(
        response: AzureDiscoveryResponse, options
    ) -> "VisualizationResponse":
        return VisualizationResponse(
            html_path="out.html",
            nodes=len(response.nodes),
            relationships=len(response.relationships),
        )

    from unittest.mock import AsyncMock

    mock_defender = AsyncMock()
    # Return one alert node so defender phase builds relationships and merges (covers 269-291)
    alert_node = ResourceNode(
        id="/subscriptions/sub-a/providers/Microsoft.Security/alerts/alert1",
        name="alert1",
        type="Microsoft.Security/alerts",
        subscription_id="sub-a",
        properties={"affectedResources": ["/subscriptions/sub-a/resourceGroups/rg/providers/Microsoft.Compute/virtualMachines/vm1"]},
    )
    mock_defender.enumerate_security_alerts = AsyncMock(return_value=[alert_node])
    mock_defender.enumerate_security_assessments = AsyncMock(return_value=[])
    mock_defender.enumerate_secure_scores = AsyncMock(return_value=[])
    mock_defender.close = AsyncMock(return_value=None)

    def fake_defender_enum(credential, base_url):
        return mock_defender

    monkeypatch.setattr(
        "azure_discovery.orchestrator.enumerate_azure_resources",
        fake_enumerate_azure,
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.render_visualization",
        fake_render,
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.emit_console_summary",
        lambda resp: None,
    )
    monkeypatch.setattr(
        "azure_discovery.orchestrator.DefenderEnumerator",
        fake_defender_enum,
    )

    out = await run_discovery(request)

    assert out.metrics is not None
    phase_names = [p.name for p in out.metrics.phases]
    assert "defender.enumeration" in phase_names
    assert len(out.nodes) >= 1


@pytest.mark.asyncio
async def test_run_discovery_rejects_none_request() -> None:
    """run_discovery raises ValueError when request is None (covers guard)."""
    with pytest.raises(ValueError, match="request cannot be None"):
        await run_discovery(None)  # type: ignore[arg-type]



"""HTML visualization helpers."""

from __future__ import annotations

import json
from datetime import datetime

from pyvis.network import Network

from ..adt_types import (
    AzureDiscoveryResponse,
    VisualizationOptions,
    VisualizationResponse,
)
from ..adt_types.errors import VisualizationError


def render_visualization(
    response: AzureDiscoveryResponse,
    options: VisualizationOptions,
) -> VisualizationResponse:
    """Render an interactive dependency graph."""

    if not response:
        raise VisualizationError("Discovery response missing")
    output_dir = options.output_dir.expanduser()
    output_dir.mkdir(parents=True, exist_ok=True)
    file_name = (
        options.file_name
        or f"azure-graph-{datetime.utcnow().strftime('%Y%m%d%H%M%S')}.html"
    )
    html_path = output_dir / file_name

    net = Network(
        height="960px",
        width="100%",
        directed=True,
        bgcolor="#05060a",
        font_color="#f5f7fa",
    )
    if options.physics_enabled:
        net.barnes_hut()
    else:
        net.force_atlas_2based(gravity=-50, central_gravity=0.015)

    for node in response.nodes:
        attributes = {
            "id": node.id,
            "type": node.type,
            "location": node.location,
            "resource_group": node.resource_group,
            "tags": node.tags,
        }
        title = json.dumps(attributes, indent=2) if options.include_attributes else node.type
        net.add_node(
            node.id,
            label=node.name,
            title=title,
            group=node.type.split("/")[-1],
        )

    for edge in response.relationships:
        net.add_edge(
            edge.source_id,
            edge.target_id,
            title=edge.relation_type,
            value=edge.weight,
        )

    try:
        net.write_html(str(html_path), notebook=False)
    except Exception as exc:  # noqa: BLE001
        raise VisualizationError("Failed to render HTML graph") from exc

    return VisualizationResponse(
        html_path=str(html_path),
        nodes=len(response.nodes),
        relationships=len(response.relationships),
    )



"""Unit tests for Purview enumerator."""

from unittest.mock import AsyncMock, MagicMock

import pytest

from azure_discovery.enumerators.purview import PurviewEnumerator
from azure_discovery.adt_types.models import PurviewConfig, ResourceNode


@pytest.fixture
def mock_credential():
    """Mock Azure credential."""
    return AsyncMock()


@pytest.fixture
def purview_config():
    """Create PurviewConfig instance."""
    return PurviewConfig(
        include_sensitivity_labels=True,
        include_dlp_policies=True,
        include_retention_policies=True,
        include_retention_labels=True,
        include_information_protection_policies=True,
        sensitivity_max_labels=1000,
        dlp_max_policies=1000,
        retention_max_policies=1000,
        retention_max_labels=1000,
        dlp_include_disabled=False,
        retention_include_disabled=False,
    )


@pytest.fixture
def purview_enumerator(mock_credential, purview_config):
    """Create PurviewEnumerator instance."""
    return PurviewEnumerator(
        tenant_id="tenant-123",
        credential=mock_credential,
        config=purview_config,
    )


@pytest.fixture
def mock_sensitivity_label():
    """Create a mock sensitivity label."""
    label = MagicMock()
    label.id = "label-guid-123"
    label.name = "Confidential"
    label.description = "Confidential information"
    label.color = "#FF0000"
    label.sensitivity = 3
    label.is_active = True
    label.tooltip = "Apply to confidential documents"
    label.additional_data = {}
    return label


@pytest.fixture
def mock_dlp_policy():
    """Create a mock DLP policy."""
    policy = MagicMock()
    policy.id = "dlp-policy-guid-456"
    policy.name = "GDPR Compliance Policy"
    policy.description = "Protect GDPR sensitive data"
    policy.state = "enabled"
    policy.mode = "Enforce"
    policy.priority = 1
    
    # Mock locations
    location1 = MagicMock()
    location1.location = "https://contoso.sharepoint.com/sites/hr"
    location1.additional_data = {"locationType": "SharePoint"}
    
    location2 = MagicMock()
    location2.location = "All"
    location2.additional_data = {"locationType": "Teams"}
    
    policy.locations = [location1, location2]
    policy.additional_data = {}
    return policy


@pytest.fixture
def mock_retention_policy():
    """Create a mock retention policy."""
    policy = MagicMock()
    policy.id = "retention-policy-guid-789"
    policy.display_name = "7 Year Retention Policy"
    policy.description = "Retain records for 7 years"
    policy.is_enabled = True
    policy.retention_duration = 2555  # days
    
    # Mock locations
    location = MagicMock()
    location.location = "All"
    location.additional_data = {"locationType": "OneDrive"}
    
    policy.locations = [location]
    policy.additional_data = {}
    return policy


@pytest.fixture
def mock_retention_label():
    """Create a mock retention label."""
    label = MagicMock()
    label.id = "retention-label-guid-abc"
    label.display_name = "Permanent Record"
    label.description = "Keep permanently"
    label.is_in_use = True
    label.retention_duration = "Unlimited"
    label.additional_data = {}
    return label


@pytest.fixture
def mock_ip_policy():
    """Create a mock information protection policy."""
    policy = MagicMock()
    policy.id = "ip-policy-guid-def"
    policy.display_name = "Default Information Protection Policy"
    policy.description = "Organization-wide IP policy"
    policy.additional_data = {}
    return policy


def _make_sensitivity_labels_mock_client(mock_response):
    """Build a mock Graph client that returns the given response from sensitivity labels get()."""
    mock_client = MagicMock()
    mock_client.security.data_security_and_governance.sensitivity_labels.get = AsyncMock(
        return_value=mock_response
    )
    return mock_client


@pytest.mark.asyncio
async def test_enumerate_sensitivity_labels_success(purview_enumerator, mock_sensitivity_label):
    """Test successful enumeration of sensitivity labels."""
    mock_response = MagicMock()
    mock_response.value = [mock_sensitivity_label]
    purview_enumerator.client = _make_sensitivity_labels_mock_client(mock_response)

    labels = await purview_enumerator.enumerate_sensitivity_labels()

    assert len(labels) == 1
    label_node = labels[0]
    assert isinstance(label_node, ResourceNode)
    assert label_node.type == "Microsoft.Purview/sensitivityLabels"
    assert label_node.name == "Confidential"
    assert label_node.properties["description"] == "Confidential information"
    assert label_node.properties["color"] == "#FF0000"
    assert label_node.properties["sensitivity"] == 3
    assert label_node.properties["isActive"] is True
    assert "labelId" in label_node.tags


@pytest.mark.asyncio
async def test_enumerate_sensitivity_labels_with_filter(purview_enumerator, mock_sensitivity_label):
    """Test sensitivity label enumeration with name filter."""
    purview_enumerator.config.sensitivity_label_filter = ["Confidential"]

    public_label = MagicMock()
    public_label.id = "label-public-123"
    public_label.name = "Public"
    public_label.description = "Public information"
    public_label.color = "#00FF00"
    public_label.sensitivity = 0
    public_label.is_active = True
    public_label.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_sensitivity_label, public_label]
    purview_enumerator.client = _make_sensitivity_labels_mock_client(mock_response)

    labels = await purview_enumerator.enumerate_sensitivity_labels()

    assert len(labels) == 1
    assert labels[0].name == "Confidential"


@pytest.mark.asyncio
async def test_enumerate_sensitivity_labels_max_limit(purview_enumerator, mock_sensitivity_label):
    """Test sensitivity label enumeration respects max limit."""
    purview_enumerator.config.sensitivity_max_labels = 2

    mock_response = MagicMock()
    mock_response.value = [mock_sensitivity_label] * 5
    purview_enumerator.client = _make_sensitivity_labels_mock_client(mock_response)

    labels = await purview_enumerator.enumerate_sensitivity_labels()

    assert len(labels) == 2


@pytest.mark.asyncio
async def test_enumerate_dlp_policies_success(purview_enumerator, mock_dlp_policy):
    """Test successful enumeration of DLP policies."""
    mock_dlp_policy.is_enabled = True
    mock_response = MagicMock()
    mock_response.value = [mock_dlp_policy]
    mock_client = MagicMock()
    mock_client.information_protection.data_loss_prevention_policies.get = AsyncMock(
        return_value=mock_response
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_dlp_policies()

    assert len(policies) == 1
    policy_node = policies[0]
    assert isinstance(policy_node, ResourceNode)
    assert policy_node.type == "Microsoft.Purview/dlpPolicies"
    assert policy_node.name == "GDPR Compliance Policy"
    assert policy_node.properties["description"] == "Protect GDPR sensitive data"
    assert policy_node.properties["isEnabled"] is True
    assert policy_node.properties["mode"] == "Enforce"
    assert policy_node.properties["priority"] == 1
    assert "policyId" in policy_node.tags


@pytest.mark.asyncio
async def test_enumerate_dlp_policies_with_filter(purview_enumerator, mock_dlp_policy):
    """Test DLP policy enumeration with name filter."""
    purview_enumerator.config.dlp_policy_filter = ["GDPR"]

    hipaa_policy = MagicMock()
    hipaa_policy.id = "dlp-hipaa-123"
    hipaa_policy.name = "HIPAA Compliance Policy"
    hipaa_policy.description = "Protect healthcare data"
    hipaa_policy.is_enabled = True
    hipaa_policy.locations = []
    hipaa_policy.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_dlp_policy, hipaa_policy]
    mock_client = MagicMock()
    mock_client.information_protection.data_loss_prevention_policies.get = AsyncMock(
        return_value=mock_response
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_dlp_policies()

    assert len(policies) == 1
    assert policies[0].name == "GDPR Compliance Policy"


@pytest.mark.asyncio
async def test_enumerate_dlp_policies_exclude_disabled(purview_enumerator, mock_dlp_policy):
    """Test DLP policy enumeration excludes disabled policies."""
    mock_dlp_policy.is_enabled = True
    disabled_policy = MagicMock()
    disabled_policy.id = "dlp-disabled-123"
    disabled_policy.name = "Disabled Policy"
    disabled_policy.description = "This is disabled"
    disabled_policy.is_enabled = False
    disabled_policy.locations = []
    disabled_policy.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_dlp_policy, disabled_policy]
    mock_client = MagicMock()
    mock_client.information_protection.data_loss_prevention_policies.get = AsyncMock(
        return_value=mock_response
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_dlp_policies()

    assert len(policies) == 1
    assert policies[0].properties["isEnabled"] is True


@pytest.mark.asyncio
async def test_enumerate_dlp_policies_include_disabled(purview_enumerator, mock_dlp_policy):
    """Test DLP policy enumeration includes disabled policies when configured."""
    purview_enumerator.config.dlp_include_disabled = True
    mock_dlp_policy.is_enabled = True

    disabled_policy = MagicMock()
    disabled_policy.id = "dlp-disabled-123"
    disabled_policy.name = "Disabled Policy"
    disabled_policy.description = "This is disabled"
    disabled_policy.is_enabled = False
    disabled_policy.locations = []
    disabled_policy.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_dlp_policy, disabled_policy]
    mock_client = MagicMock()
    mock_client.information_protection.data_loss_prevention_policies.get = AsyncMock(
        return_value=mock_response
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_dlp_policies()

    assert len(policies) == 2


@pytest.mark.asyncio
async def test_enumerate_retention_policies_success(purview_enumerator, mock_retention_policy):
    """Test successful enumeration of retention policies."""
    mock_response = MagicMock()
    mock_response.value = [mock_retention_policy]
    mock_client = MagicMock()
    mock_client.security.retention_policies.get = AsyncMock(return_value=mock_response)
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_retention_policies()

    assert len(policies) == 1
    policy_node = policies[0]
    assert isinstance(policy_node, ResourceNode)
    assert policy_node.type == "Microsoft.Purview/retentionPolicies"
    assert policy_node.name == "7 Year Retention Policy"
    assert policy_node.properties["description"] == "Retain records for 7 years"
    assert policy_node.properties["isEnabled"] is True
    assert policy_node.properties["retentionDuration"] == 2555
    assert "policyId" in policy_node.tags


@pytest.mark.asyncio
async def test_enumerate_retention_policies_with_filter(purview_enumerator, mock_retention_policy):
    """Test retention policy enumeration with name filter."""
    purview_enumerator.config.retention_policy_filter = ["7 Year"]

    short_policy = MagicMock()
    short_policy.id = "retention-short-123"
    short_policy.display_name = "30 Day Retention"
    short_policy.description = "Short retention"
    short_policy.is_enabled = True
    short_policy.locations = []
    short_policy.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_retention_policy, short_policy]
    mock_client = MagicMock()
    mock_client.security.retention_policies.get = AsyncMock(return_value=mock_response)
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_retention_policies()

    assert len(policies) == 1
    assert policies[0].name == "7 Year Retention Policy"


@pytest.mark.asyncio
async def test_enumerate_retention_labels_success(purview_enumerator, mock_retention_label):
    """Test successful enumeration of retention labels."""
    mock_response = MagicMock()
    mock_response.value = [mock_retention_label]
    mock_client = MagicMock()
    mock_client.security.labels.retention_labels.get = AsyncMock(return_value=mock_response)
    purview_enumerator.client = mock_client

    labels = await purview_enumerator.enumerate_retention_labels()

    assert len(labels) == 1
    label_node = labels[0]
    assert isinstance(label_node, ResourceNode)
    assert label_node.type == "Microsoft.Purview/retentionLabels"
    assert label_node.name == "Permanent Record"
    assert label_node.properties["description"] == "Keep permanently"
    assert label_node.properties["isInUse"] is True
    assert label_node.properties["retentionDuration"] == "Unlimited"
    assert "labelId" in label_node.tags


@pytest.mark.asyncio
async def test_enumerate_retention_labels_with_filter(purview_enumerator, mock_retention_label):
    """Test retention label enumeration with name filter."""
    purview_enumerator.config.retention_label_filter = ["Permanent"]

    temp_label = MagicMock()
    temp_label.id = "retention-temp-123"
    temp_label.display_name = "Temporary"
    temp_label.description = "Delete after use"
    temp_label.is_in_use = False
    temp_label.additional_data = {}

    mock_response = MagicMock()
    mock_response.value = [mock_retention_label, temp_label]
    mock_client = MagicMock()
    mock_client.security.labels.retention_labels.get = AsyncMock(return_value=mock_response)
    purview_enumerator.client = mock_client

    labels = await purview_enumerator.enumerate_retention_labels()

    assert len(labels) == 1
    assert labels[0].name == "Permanent Record"


@pytest.mark.asyncio
async def test_enumerate_information_protection_policies_success(purview_enumerator, mock_ip_policy):
    """Test successful enumeration of information protection policies."""
    mock_client = MagicMock()
    mock_client.information_protection.policy.get = AsyncMock(return_value=mock_ip_policy)
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_information_protection_policies()

    assert len(policies) == 1
    policy_node = policies[0]
    assert isinstance(policy_node, ResourceNode)
    assert policy_node.type == "Microsoft.Purview/informationProtectionPolicies"
    assert policy_node.name == "Default Information Protection Policy"
    assert policy_node.properties["description"] == "Organization-wide IP policy"
    assert "policyId" in policy_node.tags


@pytest.mark.asyncio
async def test_enumerate_empty_response(purview_enumerator):
    """Test enumeration with empty API response."""
    mock_response = MagicMock()
    mock_response.value = []
    purview_enumerator.client = _make_sensitivity_labels_mock_client(mock_response)

    labels = await purview_enumerator.enumerate_sensitivity_labels()
    assert len(labels) == 0


@pytest.mark.asyncio
async def test_enumerate_sensitivity_labels_api_error(purview_enumerator):
    """Test sensitivity label enumeration handles API errors gracefully."""
    mock_client = MagicMock()
    mock_client.security.data_security_and_governance.sensitivity_labels.get = AsyncMock(
        side_effect=Exception("Graph API error")
    )
    purview_enumerator.client = mock_client

    labels = await purview_enumerator.enumerate_sensitivity_labels()
    assert len(labels) == 0


@pytest.mark.asyncio
async def test_enumerate_dlp_policies_api_error(purview_enumerator):
    """Test DLP policy enumeration handles API errors gracefully."""
    mock_client = MagicMock()
    mock_client.information_protection.data_loss_prevention_policies.get = AsyncMock(
        side_effect=Exception("Graph API error")
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_dlp_policies()
    assert len(policies) == 0


@pytest.mark.asyncio
async def test_enumerate_retention_policies_api_error(purview_enumerator):
    """Test retention policy enumeration handles API errors gracefully."""
    mock_client = MagicMock()
    mock_client.security.retention_policies.get = AsyncMock(
        side_effect=Exception("Graph API error")
    )
    purview_enumerator.client = mock_client

    policies = await purview_enumerator.enumerate_retention_policies()
    assert len(policies) == 0


@pytest.mark.asyncio
async def test_purview_config_defaults():
    """Test PurviewConfig default values."""
    config = PurviewConfig()
    
    assert config.include_sensitivity_labels is True
    assert config.include_dlp_policies is True
    assert config.include_retention_policies is True
    assert config.include_retention_labels is True
    assert config.include_information_protection_policies is True
    assert config.sensitivity_max_labels == 1000
    assert config.dlp_max_policies == 1000
    assert config.retention_max_policies == 1000
    assert config.retention_max_labels == 1000
    assert config.dlp_include_disabled is False
    assert config.retention_include_disabled is False
    assert config.include_label_assignments is True
    assert config.include_policy_assignments is True



"""Entra ID discovery via Microsoft Graph."""

from __future__ import annotations

import asyncio
from typing import Any, Iterable, List, Optional, Tuple

from azure.core.credentials import TokenCredential

from ..adt_types import AzureDiscoveryRequest, ResourceNode, ResourceRelationship
from ..utils import build_environment_config
from ..utils.graph_relationships import build_graph_appid_edges
from ..utils.graph_mappers import (
    application_to_node,
    conditional_access_policy_to_node,
    domain_to_node,
    group_to_node,
    organization_to_node,
    risky_user_to_node,
    service_principal_to_node,
    user_to_node,
)
from ..utils.logging import get_logger

LOGGER = get_logger()


def _graph_id(kind: str, object_id: str) -> str:
    return f"graph://{kind}/{object_id}" if object_id else f"graph://{kind}/unknown"


def _safe_str(value: Any) -> str | None:
    if value is None:
        return None
    return str(value)


def _infer_kind_and_type(obj: Any) -> tuple[str, str]:
    """Best-effort inference of Graph object kind for ID prefixing."""

    odata_type = (
        getattr(obj, "odata_type", None)
        or getattr(obj, "odataType", None)
        or getattr(obj, "@odata.type", None)
    )
    if isinstance(odata_type, str):
        lowered = odata_type.lower()
        if "user" in lowered:
            return "user", "Microsoft.Graph/User"
        if "group" in lowered:
            return "group", "Microsoft.Graph/Group"
        if "serviceprincipal" in lowered:
            return "servicePrincipal", "Microsoft.Graph/ServicePrincipal"
        if "application" in lowered:
            return "application", "Microsoft.Graph/Application"

    # Fall back to attribute heuristics
    if getattr(obj, "user_principal_name", None) or getattr(obj, "userPrincipalName", None):
        return "user", "Microsoft.Graph/User"
    if getattr(obj, "group_types", None) or getattr(obj, "groupTypes", None):
        return "group", "Microsoft.Graph/Group"
    if getattr(obj, "app_id", None) or getattr(obj, "appId", None):
        # appId exists on both applications and service principals; if it looks like an SP,
        # callers should pass the right kind.
        return "directoryObject", "Microsoft.Graph/DirectoryObject"

    return "directoryObject", "Microsoft.Graph/DirectoryObject"


def _directory_object_to_node(obj: Any) -> ResourceNode:
    """Map a directoryObject-ish thing to a node.

    We use graph_mappers when possible, but gracefully degrade to a stub node.
    """

    kind, node_type = _infer_kind_and_type(obj)

    if node_type == "Microsoft.Graph/User":
        return user_to_node(obj)
    if node_type == "Microsoft.Graph/Group":
        return group_to_node(obj)
    if node_type == "Microsoft.Graph/ServicePrincipal":
        return service_principal_to_node(obj)
    if node_type == "Microsoft.Graph/Application":
        return application_to_node(obj)

    object_id = _safe_str(getattr(obj, "id", None)) or "unknown"
    name = (
        _safe_str(getattr(obj, "display_name", None))
        or _safe_str(getattr(obj, "displayName", None))
        or object_id
    )

    return ResourceNode(
        id=_graph_id(kind, object_id),
        name=name,
        type=node_type,
        subscription_id="Tenant",
        tags={"graph_id": object_id},
    )


async def get_all_pages(
    client: Any,
    request_builder: Any,
    *,
    operation_name: str,
    max_objects: int = 0,
    throttle_delay_seconds: float = 0.0,
) -> list[Any]:
    """Fetch all pages from a Graph collection.

    This is intentionally defined at module scope so unit tests can patch it.

    If `max_objects` is non-zero, results are truncated to that count.
    """

    from msgraph_core.tasks.page_iterator import PageIterator  # noqa: PLC0415

    results: list[Any] = []
    try:
        response = await request_builder.get()
        if response:
            iterator = PageIterator(
                response,
                client.request_adapter,
                lambda item: results.append(item) or True,
            )
            await iterator.iterate_async()
    except Exception as exc:  # noqa: BLE001
        LOGGER.warning(
            "Failed to fetch Graph collection",
            extra={
                "context": {
                    "operation": operation_name,
                    "error": str(exc),
                }
            },
        )

    if throttle_delay_seconds:
        await asyncio.sleep(float(throttle_delay_seconds))

    if max_objects:
        return results[:max_objects]
    return results


def _extend_unique_nodes(out: list[ResourceNode], new_nodes: Iterable[ResourceNode]) -> None:
    existing = {n.id for n in out}
    for n in new_nodes:
        if n.id in existing:
            continue
        existing.add(n.id)
        out.append(n)


async def enumerate_entra_resources(
    request: AzureDiscoveryRequest,
    credential: Optional[TokenCredential] = None,
) -> Tuple[List[ResourceNode], List[ResourceRelationship]]:
    """Enumerate Entra ID resources via Microsoft Graph."""

    if not request:
        raise ValueError("request cannot be None")

    if credential is None:
        raise ValueError("credential is required")

    from msgraph import GraphServiceClient  # noqa: PLC0415

    env_cfg = build_environment_config(request.environment)
    scopes = [f"{env_cfg.graph_endpoint}/.default"]
    client = GraphServiceClient(credentials=credential, scopes=scopes)

    nodes: List[ResourceNode] = []
    relationships: List[ResourceRelationship] = []

    remaining_total = request.graph_total_max_objects

    def _cap_for_collection(per_collection: int) -> int:
        if remaining_total and per_collection:
            return min(per_collection, remaining_total)
        if remaining_total:
            return remaining_total
        return per_collection

    # Organization (tenant root)
    org_node: ResourceNode | None = None
    if request.entra_include_organization and (not remaining_total or remaining_total > 0):
        org_items = await get_all_pages(
            client,
            client.organization,
            operation_name="entra.organization",
            max_objects=_cap_for_collection(1),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        if org_items:
            org_node = organization_to_node(org_items[0], tenant_id=request.tenant_id)
            nodes.append(org_node)
            if remaining_total:
                remaining_total = max(0, remaining_total - 1)

    # Domains
    if request.entra_include_domains and (not remaining_total or remaining_total > 0):
        domain_items = await get_all_pages(
            client,
            client.domains,
            operation_name="entra.domains",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        domain_nodes = [domain_to_node(d) for d in domain_items]
        _extend_unique_nodes(nodes, domain_nodes)
        if remaining_total:
            remaining_total = max(0, remaining_total - len(domain_items))

        if request.include_relationships and org_node:
            for dn in domain_nodes:
                relationships.append(
                    ResourceRelationship(
                        source_id=org_node.id,
                        target_id=dn.id,
                        relation_type="has_domain",
                        weight=1.0,
                    )
                )

    # Users
    if request.entra_include_users and (not remaining_total or remaining_total > 0):
        user_items = await get_all_pages(
            client,
            client.users,
            operation_name="entra.users",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        _extend_unique_nodes(nodes, (user_to_node(u) for u in user_items))
        if remaining_total:
            remaining_total = max(0, remaining_total - len(user_items))

    # Groups
    group_items: list[Any] = []
    group_nodes: list[ResourceNode] = []
    if request.entra_include_groups and (not remaining_total or remaining_total > 0):
        group_items = await get_all_pages(
            client,
            client.groups,
            operation_name="entra.groups",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        group_nodes = [group_to_node(g) for g in group_items]
        _extend_unique_nodes(nodes, group_nodes)
        if remaining_total:
            remaining_total = max(0, remaining_total - len(group_items))

    # Applications
    app_items: list[Any] = []
    app_nodes: list[ResourceNode] = []
    if request.entra_include_applications and (not remaining_total or remaining_total > 0):
        app_items = await get_all_pages(
            client,
            client.applications,
            operation_name="entra.applications",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        app_nodes = [application_to_node(a) for a in app_items]
        _extend_unique_nodes(nodes, app_nodes)
        if remaining_total:
            remaining_total = max(0, remaining_total - len(app_items))

    # Service principals (needed for appId correlation edges; guarded by SP expansion caps)
    sp_items: list[Any] = []
    if (
        request.entra_sp_ownership_max_sps
        and request.entra_sp_ownership_max_owners_per_sp
        and (not remaining_total or remaining_total > 0)
    ):
        sp_items = await get_all_pages(
            client,
            client.service_principals,
            operation_name="entra.service_principals",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        _extend_unique_nodes(nodes, (service_principal_to_node(sp) for sp in sp_items))
        if remaining_total:
            remaining_total = max(0, remaining_total - len(sp_items))

    # Conditional access policies
    if request.entra_include_conditional_access_policies and (not remaining_total or remaining_total > 0):
        policy_items = await get_all_pages(
            client,
            client.identity.conditional_access.policies,
            operation_name="entra.conditional_access_policies",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        _extend_unique_nodes(nodes, (conditional_access_policy_to_node(p) for p in policy_items))
        if remaining_total:
            remaining_total = max(0, remaining_total - len(policy_items))

    # Risky users
    if request.entra_include_risky_users and (not remaining_total or remaining_total > 0):
        risky_items = await get_all_pages(
            client,
            client.identity_protection.risky_users,
            operation_name="entra.risky_users",
            max_objects=_cap_for_collection(request.entra_max_objects),
            throttle_delay_seconds=request.throttle_delay_seconds,
        )
        _extend_unique_nodes(nodes, (risky_user_to_node(u) for u in risky_items))
        if remaining_total:
            remaining_total = max(0, remaining_total - len(risky_items))

    # Relationship expansions (bounded)
    if request.include_relationships:
        # 1) Group membership expansion
        if request.entra_group_membership_max_groups and request.entra_group_membership_max_members_per_group:
            for g in group_items[: request.entra_group_membership_max_groups]:
                group_id = _safe_str(getattr(g, "id", None))
                if not group_id:
                    continue
                members = await get_all_pages(
                    client,
                    client.groups.by_group_id(group_id).members,
                    operation_name="entra.group_members",
                    max_objects=request.entra_group_membership_max_members_per_group,
                    throttle_delay_seconds=request.throttle_delay_seconds,
                )
                member_nodes = [_directory_object_to_node(m) for m in members if getattr(m, "id", None)]
                _extend_unique_nodes(nodes, member_nodes)
                for mn in member_nodes:
                    relationships.append(
                        ResourceRelationship(
                            source_id=_graph_id("group", group_id),
                            target_id=mn.id,
                            relation_type="has_member",
                            weight=1.0,
                        )
                    )

        # 2) Application ownership expansion
        if request.entra_ownership_max_apps and request.entra_ownership_max_owners_per_app:
            for a in app_items[: request.entra_ownership_max_apps]:
                app_object_id = _safe_str(getattr(a, "id", None))
                if not app_object_id:
                    continue
                owners = await get_all_pages(
                    client,
                    client.applications.by_application_id(app_object_id).owners,
                    operation_name="entra.application_owners",
                    max_objects=request.entra_ownership_max_owners_per_app,
                    throttle_delay_seconds=request.throttle_delay_seconds,
                )
                owner_nodes = [_directory_object_to_node(o) for o in owners if getattr(o, "id", None)]
                _extend_unique_nodes(nodes, owner_nodes)
                for on in owner_nodes:
                    relationships.append(
                        ResourceRelationship(
                            source_id=_graph_id("application", app_object_id),
                            target_id=on.id,
                            relation_type="has_owner",
                            weight=1.0,
                        )
                    )

        # 3) Service principal ownership expansion
        if request.entra_sp_ownership_max_sps and request.entra_sp_ownership_max_owners_per_sp:
            for sp in sp_items[: request.entra_sp_ownership_max_sps]:
                sp_object_id = _safe_str(getattr(sp, "id", None))
                if not sp_object_id:
                    continue
                owners = await get_all_pages(
                    client,
                    client.service_principals.by_service_principal_id(sp_object_id).owners,
                    operation_name="entra.service_principal_owners",
                    max_objects=request.entra_sp_ownership_max_owners_per_sp,
                    throttle_delay_seconds=request.throttle_delay_seconds,
                )
                owner_nodes = [_directory_object_to_node(o) for o in owners if getattr(o, "id", None)]
                _extend_unique_nodes(nodes, owner_nodes)
                for on in owner_nodes:
                    relationships.append(
                        ResourceRelationship(
                            source_id=_graph_id("servicePrincipal", sp_object_id),
                            target_id=on.id,
                            relation_type="has_owner",
                            weight=1.0,
                        )
                    )

        # Minimal inferred relationships: appId correlation between service principals and applications.
        relationships.extend(build_graph_appid_edges(nodes))

    LOGGER.info(
        "Entra enumeration completed",
        extra={
            "context": {
                "tenant_id": request.tenant_id,
                "nodes": len(nodes),
                "relationships": len(relationships),
            }
        },
    )

    return nodes, relationships



"""Configuration models for API security and performance."""

from typing import Optional, List, Dict, Any
from pathlib import Path
from pydantic import BaseModel, Field


class AzureADAuthConfig(BaseModel):
    """Azure AD authentication configuration."""
    tenant_id: str = Field(..., description="Azure AD tenant ID")
    client_id: str = Field(..., description="Application (client) ID")
    audience: str = Field(
        "api://azure-discovery",
        description="Expected audience claim in JWT tokens",
    )
    required_scopes: List[str] = Field(
        default_factory=lambda: ["Discovery.Read"],
        description="Required OAuth scopes",
    )


class APIKeyAuthConfig(BaseModel):
    """API key authentication configuration."""
    keys_file: Optional[Path] = Field(
        None,
        description="Path to JSON file containing API key hashes and metadata",
    )
    keys: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description="In-memory key store (for development only)",
    )


class AuthConfig(BaseModel):
    """Authentication configuration."""
    backend: str = Field(
        "none",
        description="Authentication backend: azure_ad, api_key, or none",
    )
    azure_ad: Optional[AzureADAuthConfig] = None
    api_key: Optional[APIKeyAuthConfig] = None


class RateLimitConfig(BaseModel):
    """Rate limiting configuration."""
    enabled: bool = Field(True, description="Enable rate limiting")
    requests_per_minute: int = Field(
        60,
        description="Maximum requests per minute per client",
    )
    burst_size: int = Field(
        10,
        description="Maximum burst capacity (token bucket size)",
    )
    redis_url: Optional[str] = Field(
        None,
        description="Redis URL for distributed rate limiting (optional)",
    )


class QuotaConfig(BaseModel):
    """Usage quota configuration."""
    enabled: bool = Field(False, description="Enable quota enforcement")
    default_limit: int = Field(
        1000,
        description="Default quota limit per period",
    )
    period_days: int = Field(
        30,
        description="Quota reset period in days",
    )


class AuditLogConfig(BaseModel):
    """Audit logging configuration."""
    enabled: bool = Field(True, description="Enable audit logging")
    log_file: Optional[Path] = Field(
        None,
        description="Path to audit log file (JSONL format)",
    )
    include_request_body: bool = Field(
        False,
        description="Include request bodies in audit logs",
    )
    include_response_body: bool = Field(
        False,
        description="Include response bodies in audit logs",
    )


class ScaleControlConfig(BaseModel):
    """Scale control configuration for large tenants."""
    enabled: bool = Field(True, description="Enable adaptive rate control")
    initial_rps: float = Field(
        10.0,
        description="Initial requests per second",
    )
    min_rps: float = Field(1.0, description="Minimum requests per second")
    max_rps: float = Field(50.0, description="Maximum requests per second")
    max_concurrent_batches: int = Field(
        5,
        description="Maximum concurrent batch operations",
    )
    adaptive_batching: bool = Field(
        True,
        description="Enable adaptive batch sizing",
    )
    initial_batch_size: int = Field(
        1000,
        description="Initial batch size for paginated operations",
    )


class APIConfig(BaseModel):
    """Complete API configuration."""
    auth: AuthConfig = Field(default_factory=AuthConfig)
    rate_limit: RateLimitConfig = Field(default_factory=RateLimitConfig)
    quota: QuotaConfig = Field(default_factory=QuotaConfig)
    audit_log: AuditLogConfig = Field(default_factory=AuditLogConfig)
    scale_controls: ScaleControlConfig = Field(default_factory=ScaleControlConfig)
    
    @classmethod
    def from_toml(cls, config_file: Path) -> "APIConfig":
        """Load configuration from TOML file.
        
        Args:
            config_file: Path to TOML configuration file
            
        Returns:
            APIConfig instance
        """
        import tomli
        
        with open(config_file, "rb") as f:
            data = tomli.load(f)
        
        return cls(**data.get("api", {}))



"""Tests for Azure client helpers."""

import os
from unittest.mock import MagicMock, patch

import pytest

from azure_discovery.adt_types import AzureDiscoveryRequest, AzureEnvironment
from azure_discovery.adt_types.errors import AzureClientError
from azure_discovery.utils.azure_clients import (
    _is_headless,
    build_environment_config,
    get_credential,
)


class TestBuildEnvironmentConfig:
    """Tests for build_environment_config."""

    def test_azure_public(self) -> None:
        config = build_environment_config(AzureEnvironment.AZURE_PUBLIC)
        assert config.name == AzureEnvironment.AZURE_PUBLIC
        assert "management.azure.com" in config.resource_manager
        assert "graph.microsoft.com" in config.graph_endpoint

    def test_azure_gov(self) -> None:
        config = build_environment_config(AzureEnvironment.AZURE_GOV)
        assert config.name == AzureEnvironment.AZURE_GOV
        assert "usgovcloudapi" in config.resource_manager or "gov" in config.graph_endpoint.lower()

    def test_azure_china(self) -> None:
        config = build_environment_config(AzureEnvironment.AZURE_CHINA)
        assert config.name == AzureEnvironment.AZURE_CHINA
        assert "chinacloudapi" in config.resource_manager or "china" in config.graph_endpoint.lower()

    def test_azure_germany(self) -> None:
        config = build_environment_config(AzureEnvironment.AZURE_GERMANY)
        assert config.name == AzureEnvironment.AZURE_GERMANY

    def test_azure_stack(self) -> None:
        config = build_environment_config(AzureEnvironment.AZURE_STACK)
        assert config.name == AzureEnvironment.AZURE_STACK
        assert "azurestack" in config.resource_manager.lower()

    def test_unsupported_environment_raises(self) -> None:
        with pytest.raises(AzureClientError, match="Unsupported environment"):
            build_environment_config("unsupported")  # type: ignore[arg-type]


class TestGetCredential:
    """Tests for get_credential."""

    def test_returns_chained_credential(self) -> None:
        request = AzureDiscoveryRequest(tenant_id="t1")
        config = build_environment_config(AzureEnvironment.AZURE_PUBLIC)
        cred = get_credential(request, config)
        assert cred is not None
        assert "ChainedTokenCredential" in type(cred).__name__

    @patch("azure_discovery.utils.azure_clients.AzureCliCredential")
    def test_prefer_cli_credentials_adds_cli_to_chain(
        self, mock_cli_cred: MagicMock
    ) -> None:
        """With prefer_cli_credentials=True, AzureCliCredential is included in the chain."""
        mock_cli_cred.return_value = MagicMock()
        request = AzureDiscoveryRequest(tenant_id="t1", prefer_cli_credentials=True)
        config = build_environment_config(AzureEnvironment.AZURE_PUBLIC)
        cred = get_credential(request, config)
        assert cred is not None
        mock_cli_cred.assert_called_once()


class TestIsHeadless:
    """Tests for _is_headless."""

    def test_wsl_returns_true(self) -> None:
        with patch.dict("os.environ", {"WSL_DISTRO_NAME": "Ubuntu"}, clear=False):
            assert _is_headless() is True

    def test_posix_no_display_returns_true(self) -> None:
        with patch.dict("os.environ", {}, clear=True), patch("os.name", "posix"):
            # No DISPLAY
            assert _is_headless() is True

    def test_posix_with_display_returns_false(self) -> None:
        # Simulate non-WSL Linux with DISPLAY (patch so WSL is not set, DISPLAY is).
        real_get = os.environ.get

        def mock_get(k: str, d: str | None = None) -> str | None:
            if k == "WSL_DISTRO_NAME":
                return None
            if k == "DISPLAY":
                return ":0"
            return real_get(k, d)

        with patch("azure_discovery.utils.azure_clients.os.environ.get", side_effect=mock_get), patch(
            "azure_discovery.utils.azure_clients.os.name", "posix"
        ):
            assert _is_headless() is False

    def test_wsl_with_display_still_headless(self) -> None:
        with patch.dict(
            "os.environ", {"WSL_DISTRO_NAME": "Ubuntu", "DISPLAY": ":0"}, clear=False
        ):
            assert _is_headless() is True



"""Unit tests for M365 helper functions."""

import pytest

from azure_discovery.adt_types.models import ResourceNode, ResourceRelationship
from azure_discovery.utils.m365_helpers import (
    build_m365_relationships,
    get_sharepoint_sites,
    get_teams,
    get_onedrive_drives,
    get_exchange_mailboxes,
    group_m365_by_type,
)


@pytest.fixture
def mock_sharepoint_site():
    """Create a mock SharePoint site node."""
    return ResourceNode(
        id="sharepoint://sites/site-123",
        name="Engineering Site",
        type="Microsoft.SharePoint/sites",
        subscription_id="Tenant",
        properties={
            "web_url": "https://contoso.sharepoint.com/sites/engineering",
            "description": "Engineering team collaboration",
            "owner_upn": "alice@contoso.com",
        },
        tags={"siteId": "site-123"},
    )


@pytest.fixture
def mock_team():
    """Create a mock Team node."""
    return ResourceNode(
        id="teams://teams/team-456",
        name="Product Development",
        type="Microsoft.Teams/teams",
        subscription_id="Tenant",
        properties={
            "mail": "productdev@contoso.com",
            "visibility": "Private",
            "member_upns": ["bob@contoso.com"],
        },
        tags={"teamId": "team-456"},
    )


@pytest.fixture
def mock_onedrive_drive():
    """Create a mock OneDrive drive node."""
    return ResourceNode(
        id="onedrive://drives/drive-789",
        name="Alice's OneDrive",
        type="Microsoft.OneDrive/drives",
        subscription_id="Tenant",
        properties={
            "drive_type": "business",
            "owner_upn": "alice@contoso.com",
        },
        tags={"driveId": "drive-789"},
    )


@pytest.fixture
def mock_exchange_mailbox():
    """Create a mock Exchange mailbox node."""
    return ResourceNode(
        id="exchange://mailboxes/user-dave",
        name="dave@contoso.com",
        type="Microsoft.Exchange/mailboxes",
        subscription_id="Tenant",
        properties={
            "user_principal_name": "dave@contoso.com",
            "mail": "dave@contoso.com",
        },
        tags={"userId": "user-dave"},
    )


@pytest.fixture
def mock_user():
    """Create a mock Entra user node (id must match helper's owner_id format)."""
    return ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/alice@contoso.com",
        name="Alice Anderson",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={"user_principal_name": "alice@contoso.com"},
        tags={"graph_id": "user-alice"},
    )


def test_get_sharepoint_sites(mock_sharepoint_site, mock_team):
    """Test filtering nodes to SharePoint sites."""
    nodes = [mock_sharepoint_site, mock_team]
    sites = get_sharepoint_sites(nodes)
    
    assert len(sites) == 1
    assert sites[0].type == "Microsoft.SharePoint/sites"


def test_get_teams(mock_sharepoint_site, mock_team):
    """Test filtering nodes to Teams."""
    nodes = [mock_sharepoint_site, mock_team]
    teams = get_teams(nodes)
    
    assert len(teams) == 1
    assert teams[0].type == "Microsoft.Teams/teams"


def test_get_onedrive_drives(mock_onedrive_drive, mock_team):
    """Test filtering nodes to OneDrive drives."""
    nodes = [mock_onedrive_drive, mock_team]
    drives = get_onedrive_drives(nodes)
    
    assert len(drives) == 1
    assert drives[0].type == "Microsoft.OneDrive/drives"


def test_get_exchange_mailboxes(mock_exchange_mailbox, mock_team):
    """Test filtering nodes to Exchange mailboxes."""
    nodes = [mock_exchange_mailbox, mock_team]
    mailboxes = get_exchange_mailboxes(nodes)
    
    assert len(mailboxes) == 1
    assert mailboxes[0].type == "Microsoft.Exchange/mailboxes"


def test_group_m365_by_type(mock_sharepoint_site, mock_team, mock_onedrive_drive, mock_exchange_mailbox):
    """Test grouping M365 nodes by type."""
    nodes = [mock_sharepoint_site, mock_team, mock_onedrive_drive, mock_exchange_mailbox]
    grouped = group_m365_by_type(nodes)
    
    assert "Microsoft.SharePoint/sites" in grouped
    assert "Microsoft.Teams/teams" in grouped
    assert "Microsoft.OneDrive/drives" in grouped
    assert "Microsoft.Exchange/mailboxes" in grouped
    
    assert len(grouped["Microsoft.SharePoint/sites"]) == 1
    assert len(grouped["Microsoft.Teams/teams"]) == 1
    assert len(grouped["Microsoft.OneDrive/drives"]) == 1
    assert len(grouped["Microsoft.Exchange/mailboxes"]) == 1


def test_build_sharepoint_relationships(mock_sharepoint_site, mock_user):
    """Test building SharePoint -> User relationships."""
    m365_nodes = [mock_sharepoint_site]
    all_nodes = [mock_user]
    
    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )
    
    assert len(relationships) == 1
    assert relationships[0].relation_type == "owns"
    assert relationships[0].source_id == "https://graph.microsoft.com/v1.0/users/alice@contoso.com"
    assert relationships[0].target_id == "sharepoint://sites/site-123"
    assert len(materialized) == 0


def test_build_team_relationships(mock_team, mock_user):
    """Test building Team -> User relationships."""
    bob_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/bob@contoso.com",
        name="Bob Brown",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-bob"},
    )

    m365_nodes = [mock_team]
    all_nodes = [bob_user]

    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )

    assert len(relationships) == 1
    assert relationships[0].relation_type == "member_of"
    assert relationships[0].source_id == "https://graph.microsoft.com/v1.0/users/bob@contoso.com"
    assert relationships[0].target_id == "teams://teams/team-456"


def test_build_onedrive_relationships(mock_onedrive_drive, mock_user):
    """Test building OneDrive -> User relationships."""
    m365_nodes = [mock_onedrive_drive]
    all_nodes = [mock_user]
    
    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )
    
    assert len(relationships) == 1
    assert relationships[0].relation_type == "owns"
    assert relationships[0].source_id == "https://graph.microsoft.com/v1.0/users/alice@contoso.com"
    assert relationships[0].target_id == "onedrive://drives/drive-789"


def test_build_exchange_relationships(mock_exchange_mailbox):
    """Test building Exchange -> User relationships."""
    dave_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/dave@contoso.com",
        name="Dave Davis",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-dave"},
    )

    m365_nodes = [mock_exchange_mailbox]
    all_nodes = [dave_user]

    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )

    assert len(relationships) == 1
    assert relationships[0].relation_type == "owns"
    assert relationships[0].source_id == "https://graph.microsoft.com/v1.0/users/dave@contoso.com"
    assert relationships[0].target_id == "exchange://mailboxes/user-dave"


def test_build_relationships_with_materialization(mock_sharepoint_site):
    """Test materialization of missing users."""
    m365_nodes = [mock_sharepoint_site]
    all_nodes = []

    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=True,
    )

    assert len(relationships) == 1
    assert len(materialized) == 1

    materialized_user = materialized[0]
    assert materialized_user.type == "Microsoft.Graph/users"
    assert materialized_user.id == "https://graph.microsoft.com/v1.0/users/alice@contoso.com"
    assert materialized_user.name == "alice@contoso.com"
    assert materialized_user.properties.get("materialized") is True


def test_build_relationships_without_materialization(mock_sharepoint_site):
    """Test skipping relationships when users missing and materialization disabled."""
    m365_nodes = [mock_sharepoint_site]
    all_nodes = []  # No existing users
    
    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )
    
    # Should not create relationship or materialize user
    assert len(relationships) == 0
    assert len(materialized) == 0


def test_build_relationships_all_workloads(
    mock_sharepoint_site,
    mock_team,
    mock_onedrive_drive,
    mock_exchange_mailbox,
    mock_user,
):
    """Test building relationships for all M365 workload types."""
    bob_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/bob@contoso.com",
        name="Bob Brown",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-bob"},
    )
    dave_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/dave@contoso.com",
        name="Dave Davis",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-dave"},
    )

    m365_nodes = [
        mock_sharepoint_site,
        mock_team,
        mock_onedrive_drive,
        mock_exchange_mailbox,
    ]
    all_nodes = [mock_user, bob_user, dave_user]

    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )

    assert len(relationships) == 4
    assert len(materialized) == 0

    rel_types = {r.relation_type for r in relationships}
    assert "owns" in rel_types
    assert "member_of" in rel_types


def test_build_relationships_empty_input():
    """Test building relationships with empty inputs."""
    relationships, materialized = build_m365_relationships(
        m365_nodes=[],
        all_nodes=[],
        materialize_missing=False,
    )
    
    assert len(relationships) == 0
    assert len(materialized) == 0


def test_build_relationships_no_owner_tags(mock_sharepoint_site):
    """Test handling M365 nodes without owner tags."""
    # Remove owner tags
    site_no_owner = ResourceNode(
        id="sharepoint://sites/site-orphan",
        name="Orphan Site",
        type="Microsoft.SharePoint/sites",
        subscription_id="Tenant",
        properties={},
        tags={"siteId": "site-orphan"},  # No ownerUserId
    )
    
    m365_nodes = [site_no_owner]
    all_nodes = []
    
    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=True,
    )
    
    # Should not create relationships for orphan resources
    assert len(relationships) == 0
    assert len(materialized) == 0


def test_team_multiple_owners(mock_team):
    """Test Teams with multiple owners creates multiple relationships."""
    mock_team.properties["member_upns"] = ["bob@contoso.com", "charlie@contoso.com"]
    bob_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/bob@contoso.com",
        name="Bob Brown",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-bob"},
    )
    charlie_user = ResourceNode(
        id="https://graph.microsoft.com/v1.0/users/charlie@contoso.com",
        name="Charlie Chen",
        type="Microsoft.Graph/users",
        subscription_id="Tenant",
        properties={},
        tags={"graph_id": "user-charlie"},
    )

    m365_nodes = [mock_team]
    all_nodes = [bob_user, charlie_user]

    relationships, materialized = build_m365_relationships(
        m365_nodes=m365_nodes,
        all_nodes=all_nodes,
        materialize_missing=False,
    )

    assert len(relationships) == 2
    assert all(r.relation_type == "member_of" for r in relationships)

    source_ids = {r.source_id for r in relationships}
    assert "https://graph.microsoft.com/v1.0/users/bob@contoso.com" in source_ids
    assert "https://graph.microsoft.com/v1.0/users/charlie@contoso.com" in source_ids


def test_filtering_helpers_empty_input():
    """Test filtering helpers with empty input."""
    assert len(get_sharepoint_sites([])) == 0
    assert len(get_teams([])) == 0
    assert len(get_onedrive_drives([])) == 0
    assert len(get_exchange_mailboxes([])) == 0


def test_group_m365_by_type_empty_input():
    """Test grouping with empty input."""
    grouped = group_m365_by_type([])
    assert len(grouped) == 0



"""FastAPI surface for Azure discovery."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Optional

from fastapi import Depends, FastAPI, HTTPException, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.routing import APIRouter
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

from .adt_types import AzureDiscoveryRequest, AzureDiscoveryResponse
from .auth import AuthBackend, get_auth_provider, initialize_auth
from .auth.models import AuthenticationError, UserContext
from .config.api_config import APIConfig
from .middleware.audit import AuditLogMiddleware, FileAuditLogWriter, LoggerAuditLogWriter
from .middleware.quota import InMemoryQuotaStore, QuotaMiddleware
from .middleware.rate_limit import RateLimitConfig as MiddlewareRateLimitConfig
from .middleware.rate_limit import RateLimitMiddleware
from .orchestrator import run_discovery
from .utils.config_files import deep_merge, load_config_file
from .utils.logging import get_logger

LOGGER = get_logger()

# ---------------------------------------------------------------------------
# Load API configuration
# ---------------------------------------------------------------------------

_api_config_path = os.getenv("AZURE_DISCOVERY_API_CONFIG")
_api_config = APIConfig()
if _api_config_path:
    try:
        _api_config = APIConfig.from_toml(Path(_api_config_path))
        LOGGER.info(
            "Loaded API configuration",
            extra={"context": {"path": _api_config_path}},
        )
    except Exception as exc:  # noqa: BLE001
        LOGGER.error(
            "Failed to load AZURE_DISCOVERY_API_CONFIG; using defaults",
            extra={"context": {"path": _api_config_path, "error": str(exc)}},
        )

# ---------------------------------------------------------------------------
# Initialize authentication
# ---------------------------------------------------------------------------

_auth_backend = AuthBackend(_api_config.auth.backend)
_auth_init_config: dict = {}

if _auth_backend == AuthBackend.AZURE_AD and _api_config.auth.azure_ad:
    _auth_init_config = {
        "tenant_id": _api_config.auth.azure_ad.tenant_id,
        "client_id": _api_config.auth.azure_ad.client_id,
        "audience": _api_config.auth.azure_ad.audience,
    }
elif _auth_backend == AuthBackend.API_KEY and _api_config.auth.api_key:
    _auth_init_config = {"keys": _api_config.auth.api_key.keys}

initialize_auth(_auth_backend, _auth_init_config)

# ---------------------------------------------------------------------------
# Application setup
# ---------------------------------------------------------------------------

app = FastAPI(title="Azure Discovery", version="1.0.0")

# CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=os.getenv("AZURE_DISCOVERY_CORS_ORIGINS", "*").split(","),
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Audit logging middleware
if _api_config.audit_log.enabled:
    if _api_config.audit_log.log_file:
        _audit_writer = FileAuditLogWriter(_api_config.audit_log.log_file)
    else:
        _audit_writer = LoggerAuditLogWriter()
    app.add_middleware(
        AuditLogMiddleware,
        writer=_audit_writer,
        include_request_body=_api_config.audit_log.include_request_body,
        include_response_body=_api_config.audit_log.include_response_body,
    )

# Quota middleware
if _api_config.quota.enabled:
    _quota_store = InMemoryQuotaStore(
        default_limit=_api_config.quota.default_limit,
        period_days=_api_config.quota.period_days,
    )
    app.add_middleware(QuotaMiddleware, quota_store=_quota_store)

# Rate limiting middleware
if _api_config.rate_limit.enabled:
    _rate_limit_config = MiddlewareRateLimitConfig(
        enabled=_api_config.rate_limit.enabled,
        requests_per_minute=_api_config.rate_limit.requests_per_minute,
        burst_size=_api_config.rate_limit.burst_size,
        redis_url=_api_config.rate_limit.redis_url,
    )
    app.add_middleware(RateLimitMiddleware, config=_rate_limit_config)

# ---------------------------------------------------------------------------
# Authentication dependency
# ---------------------------------------------------------------------------

security = HTTPBearer(auto_error=False)


async def authenticate_request(
    credentials: Optional[HTTPAuthorizationCredentials] = Security(security),
) -> Optional[UserContext]:
    """Dependency for route authentication.

    Returns None when auth backend is 'none' (development mode).
    Raises 401 for all other backends when credentials are missing or invalid.
    """
    auth_provider = get_auth_provider()

    # No auth configured (development mode)
    if auth_provider is None:
        return None

    if not credentials:
        raise HTTPException(status_code=401, detail="Missing authentication credentials")

    try:
        user_context = await auth_provider.authenticate(credentials.credentials)
        return UserContext(**user_context)
    except AuthenticationError as exc:
        raise HTTPException(status_code=401, detail=str(exc))
    except Exception:
        raise HTTPException(status_code=401, detail="Invalid credentials")


# ---------------------------------------------------------------------------
# Request defaults from config file
# ---------------------------------------------------------------------------

_DEFAULT_REQUEST_OVERRIDES: dict = {}
_default_config_path = os.getenv("AZURE_DISCOVERY_CONFIG")
if _default_config_path:
    try:
        _DEFAULT_REQUEST_OVERRIDES = load_config_file(Path(_default_config_path))
        LOGGER.info(
            "Loaded AzureDiscoveryRequest defaults from config",
            extra={"context": {"path": _default_config_path}},
        )
    except Exception as exc:  # noqa: BLE001
        LOGGER.error(
            "Failed to load AZURE_DISCOVERY_CONFIG; continuing without defaults",
            extra={"context": {"path": _default_config_path, "error": str(exc)}},
        )

# ---------------------------------------------------------------------------
# API v1 versioned router
# ---------------------------------------------------------------------------

v1_router = APIRouter(prefix="/v1", tags=["v1"])


@v1_router.post(
    "/discover",
    response_model=AzureDiscoveryResponse,
    tags=["discovery"],
)
async def v1_discover_endpoint(
    request: AzureDiscoveryRequest,
    user: Optional[UserContext] = Depends(authenticate_request),
) -> AzureDiscoveryResponse:
    """Versioned discovery endpoint (v1)."""
    return await _run_discovery_with_defaults(request)


@v1_router.get("/healthz", tags=["system"])
async def v1_health_check() -> dict:
    return {"status": "ok", "version": "v1"}


app.include_router(v1_router)

# ---------------------------------------------------------------------------
# Root (unversioned) endpoints — backward compatible
# ---------------------------------------------------------------------------


@app.get("/healthz", tags=["system"])
async def health_check() -> dict:
    return {"status": "ok"}


@app.post(
    "/discover",
    response_model=AzureDiscoveryResponse,
    tags=["discovery"],
)
async def discover_endpoint(
    request: AzureDiscoveryRequest,
    user: Optional[UserContext] = Depends(authenticate_request),
) -> AzureDiscoveryResponse:
    return await _run_discovery_with_defaults(request)


@app.get("/visuals/{file_name}", tags=["visualization"])
async def visualization_endpoint(file_name: str) -> FileResponse:
    if not file_name:
        raise HTTPException(status_code=400, detail="file_name missing")
    html_path = Path("artifacts/graphs") / file_name
    if not html_path.is_file():
        raise HTTPException(status_code=404, detail="Visualization not found")
    return FileResponse(html_path)


# ---------------------------------------------------------------------------
# Shared helpers
# ---------------------------------------------------------------------------


async def _run_discovery_with_defaults(
    request: AzureDiscoveryRequest,
) -> AzureDiscoveryResponse:
    """Apply default overrides and run discovery."""
    if _DEFAULT_REQUEST_OVERRIDES:
        incoming = request.model_dump(exclude_unset=True)
        merged = deep_merge(_DEFAULT_REQUEST_OVERRIDES, incoming)
        request = AzureDiscoveryRequest.model_validate(merged)

    LOGGER.info(
        "API discovery invoked",
        extra={"context": {"tenant_id": request.tenant_id}},
    )
    return await run_discovery(request)



"""Tests for logging configuration."""

import logging

from azure_discovery.utils.logging import get_logger


def test_get_logger_quiet_sets_error_level_when_handlers_present() -> None:
    """When logger already has handlers and quiet=True, level is set to ERROR."""
    # First call attaches NullHandler so logger.handlers is truthy
    get_logger()
    # Second call with quiet=True hits the branch that sets ERROR and returns
    logger = get_logger(quiet=True)
    assert logger.level == logging.ERROR


def test_get_logger_sets_level_from_env_when_no_handlers() -> None:
    """When logger has no handlers, level is set from AZURE_DISCOVERY_LOG_LEVEL (or INFO)."""
    logger = logging.getLogger("azure_discovery")
    logger.handlers.clear()
    logger.setLevel(logging.NOTSET)
    # get_logger(quiet=False) with no handlers adds NullHandler and sets level from env
    out = get_logger(quiet=False)
    assert out.level != logging.NOTSET


def test_get_logger_quiet_sets_error_when_no_handlers() -> None:
    """When logger has no handlers and quiet=True, level is set to ERROR (covers line 62)."""
    logger = logging.getLogger("azure_discovery")
    logger.handlers.clear()
    out = get_logger(quiet=True)
    assert out.level == logging.ERROR



"""Typed interfaces shared across the Azure discovery tool."""

from .errors import (
    AzureClientError,
    InvalidTargetError,
    VisualizationError,
)
from .models import (
    AzureDiscoveryRequest,
    AzureDiscoveryResponse,
    AzureEnvironment,
    AzureEnvironmentConfig,
    ComplianceControl,
    ComplianceMapping,
    DiscoveryFilter,
    DiscoveryMergeMetrics,
    DiscoveryMetrics,
    DiscoveryPhaseMetrics,
    PolicyControlEvidence,
    PolicyInitiativeInfo,
    PolicyStateEvidence,
    ResourceNode,
    ResourceRelationship,
    ScaleControlConfig,
    VisualizationOptions,
    VisualizationResponse,
)

__all__ = [
    "AzureClientError",
    "InvalidTargetError",
    "VisualizationError",
    "AzureDiscoveryRequest",
    "AzureDiscoveryResponse",
    "AzureEnvironment",
    "AzureEnvironmentConfig",
    "DiscoveryFilter",
    "DiscoveryMergeMetrics",
    "DiscoveryMetrics",
    "DiscoveryPhaseMetrics",
    "ResourceNode",
    "ResourceRelationship",
    "VisualizationOptions",
    "VisualizationResponse",
]



"""Audit logging middleware for compliance and security."""

import time
import json
import logging
from abc import ABC, abstractmethod
from typing import Dict, Any, Optional
from datetime import datetime, timezone
from pathlib import Path
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

LOGGER = logging.getLogger(__name__)


class AuditLogWriter(ABC):
    """Abstract interface for audit log writing."""
    
    @abstractmethod
    async def write(self, entry: Dict[str, Any]) -> None:
        """Write audit log entry.
        
        Args:
            entry: Audit log entry dictionary
        """
        pass


class FileAuditLogWriter(AuditLogWriter):
    """Write audit logs to file (JSONL format)."""
    
    def __init__(self, log_file: Path):
        """Initialize file audit log writer.
        
        Args:
            log_file: Path to audit log file
        """
        self.log_file = log_file
        self.log_file.parent.mkdir(parents=True, exist_ok=True)
    
    async def write(self, entry: Dict[str, Any]) -> None:
        """Write audit log entry to file.
        
        Args:
            entry: Audit log entry dictionary
        """
        with open(self.log_file, "a") as f:
            f.write(json.dumps(entry) + "\n")


class LoggerAuditLogWriter(AuditLogWriter):
    """Write audit logs using Python logging."""
    
    def __init__(self, logger_name: str = "azure_discovery.audit"):
        """Initialize logger audit log writer.
        
        Args:
            logger_name: Name of logger to use
        """
        self.logger = logging.getLogger(logger_name)
    
    async def write(self, entry: Dict[str, Any]) -> None:
        """Write audit log entry using logger.
        
        Args:
            entry: Audit log entry dictionary
        """
        self.logger.info(
            "API Request",
            extra={"audit": entry},
        )


class AuditLogMiddleware(BaseHTTPMiddleware):
    """Log all API requests for security and compliance."""
    
    def __init__(
        self,
        app,
        writer: AuditLogWriter,
        include_request_body: bool = False,
        include_response_body: bool = False,
    ):
        """Initialize audit log middleware.
        
        Args:
            app: FastAPI application
            writer: Audit log writer implementation
            include_request_body: Whether to log request bodies
            include_response_body: Whether to log response bodies
        """
        super().__init__(app)
        self.writer = writer
        self.include_request_body = include_request_body
        self.include_response_body = include_response_body
    
    async def dispatch(self, request: Request, call_next):
        """Process request with audit logging.
        
        Args:
            request: Incoming HTTP request
            call_next: Next middleware in chain
            
        Returns:
            HTTP response
        """
        start_time = time.time()
        
        # Capture request details
        audit_entry = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "method": request.method,
            "path": request.url.path,
            "query_params": dict(request.query_params),
            "client_ip": request.client.host if request.client else None,
            "user_agent": request.headers.get("user-agent"),
        }
        
        # Add authenticated user info if available
        if hasattr(request.state, "user"):
            audit_entry["user"] = {
                "id": request.state.user.id,
                "email": request.state.user.email,
                "source": request.state.user.source,
            }
        
        # Optionally capture request body
        if self.include_request_body and request.method in ["POST", "PUT", "PATCH"]:
            try:
                body = await request.body()
                # Avoid logging large bodies
                if len(body) < 10000:
                    audit_entry["request_body"] = body.decode("utf-8")
            except Exception as exc:
                LOGGER.debug(f"Failed to capture request body: {exc}")
        
        # Process request
        error: Optional[Exception] = None
        try:
            response = await call_next(request)
            audit_entry["status_code"] = response.status_code
        except Exception as exc:
            error = exc
            audit_entry["error"] = str(exc)
            audit_entry["error_type"] = type(exc).__name__
            raise
        finally:
            # Record duration
            audit_entry["duration_ms"] = (time.time() - start_time) * 1000
            
            # Write audit log
            try:
                await self.writer.write(audit_entry)
            except Exception as exc:
                LOGGER.error(f"Failed to write audit log: {exc}")
        
        return response



"""Tests for Microsoft Graph to ResourceNode mappers."""

from unittest.mock import MagicMock

import pytest

from azure_discovery.adt_types import ResourceNode
from azure_discovery.utils.graph_mappers import (
    application_to_node,
    conditional_access_policy_to_node,
    group_to_node,
    risky_user_to_node,
    service_principal_to_node,
)


def test_group_to_node_with_list_group_types() -> None:
    """group_to_node sets groupType from first element and groupTypes from list."""
    group = MagicMock()
    group.id = "g1"
    group.display_name = "Group One"
    group.mail = None
    group.group_types = ["Unified", "Dynamic"]

    node = group_to_node(group)

    assert node.type == "Microsoft.Graph/Group"
    assert node.tags["groupType"] == "Unified"
    assert node.tags["groupTypes"] == "Unified,Dynamic"


def test_group_to_node_with_non_list_group_types() -> None:
    """group_to_node sets groupType from scalar group_types."""
    group = MagicMock()
    group.id = "g2"
    group.displayName = "Group Two"
    group.mail = None
    group.group_types = None
    group.groupTypes = "Unified"

    node = group_to_node(group)

    assert node.tags["groupType"] == "Unified"


def test_application_to_node_with_app_id() -> None:
    """application_to_node includes appId in tags when present."""
    app = MagicMock()
    app.id = "app-1"
    app.display_name = "My App"
    app.app_id = "00000000-0000-0000-0000-000000000001"

    node = application_to_node(app)

    assert node.type == "Microsoft.Graph/Application"
    assert node.tags["appId"] == "00000000-0000-0000-0000-000000000001"


def test_service_principal_to_node_with_app_id() -> None:
    """service_principal_to_node includes appId in tags when present."""
    sp = MagicMock()
    sp.id = "sp-1"
    sp.displayName = "Service Principal"
    sp.app_id = None  # mapper tries app_id then appId
    sp.appId = "00000000-0000-0000-0000-000000000002"

    node = service_principal_to_node(sp)

    assert node.type == "Microsoft.Graph/ServicePrincipal"
    assert node.tags["appId"] == "00000000-0000-0000-0000-000000000002"


def test_conditional_access_policy_to_node_with_state() -> None:
    """conditional_access_policy_to_node includes state in tags when present."""
    policy = MagicMock()
    policy.id = "cap-1"
    policy.displayName = "Block Legacy"
    policy.state = "enabled"

    node = conditional_access_policy_to_node(policy)

    assert node.type == "Microsoft.Graph/ConditionalAccessPolicy"
    assert node.tags["state"] == "enabled"


def test_risky_user_to_node_with_risk_and_upn() -> None:
    """risky_user_to_node includes riskLevel, riskState, userPrincipalName in tags."""
    risky = MagicMock()
    risky.id = "ru-1"
    risky.user_display_name = "Risky User"
    risky.user_principal_name = None  # mapper tries user_principal_name then userPrincipalName
    risky.userPrincipalName = "user@contoso.com"
    risky.risk_level = "high"
    risky.risk_state = "confirmedCompromised"
    risky.riskLevel = "high"
    risky.riskState = "confirmedCompromised"

    node = risky_user_to_node(risky)

    assert node.type == "Microsoft.Graph/RiskyUser"
    assert node.tags["riskLevel"] == "high"
    assert node.tags["riskState"] == "confirmedCompromised"
    assert node.tags["userPrincipalName"] == "user@contoso.com"
    assert node.name == "Risky User"



"""Rate limiting middleware for API protection."""

import time
import logging
from collections import defaultdict
from typing import Optional, Dict, Any
from pydantic import BaseModel
from fastapi import Request, HTTPException
from starlette.middleware.base import BaseHTTPMiddleware

try:
    import redis.asyncio as aioredis
    REDIS_AVAILABLE = True
except ImportError:
    REDIS_AVAILABLE = False

LOGGER = logging.getLogger(__name__)


class RateLimitConfig(BaseModel):
    """Configuration for rate limiting."""
    enabled: bool = True
    requests_per_minute: int = 60
    burst_size: int = 10
    redis_url: Optional[str] = None


class RateLimitMiddleware(BaseHTTPMiddleware):
    """Token bucket rate limiting middleware.
    
    Implements rate limiting with configurable backend (memory or Redis).
    Uses token bucket algorithm for smooth rate control with burst capacity.
    """
    
    def __init__(
        self,
        app,
        config: RateLimitConfig,
    ):
        """Initialize rate limit middleware.
        
        Args:
            app: FastAPI application
            config: Rate limit configuration
        """
        super().__init__(app)
        self.config = config
        
        if config.redis_url and REDIS_AVAILABLE:
            self.redis = aioredis.from_url(config.redis_url)
            self.backend = "redis"
            LOGGER.info("Rate limiting using Redis backend")
        else:
            if config.redis_url and not REDIS_AVAILABLE:
                LOGGER.warning(
                    "Redis URL provided but redis package not available. "
                    "Falling back to in-memory rate limiting."
                )
            self.buckets: Dict[str, Dict[str, Any]] = defaultdict(
                lambda: {"tokens": config.burst_size, "last_update": time.time()}
            )
            self.backend = "memory"
            LOGGER.info("Rate limiting using in-memory backend")
    
    async def dispatch(self, request: Request, call_next):
        """Process request with rate limiting.
        
        Args:
            request: Incoming HTTP request
            call_next: Next middleware in chain
            
        Returns:
            HTTP response with rate limit headers
            
        Raises:
            HTTPException: 429 if rate limit exceeded
        """
        if not self.config.enabled:
            return await call_next(request)
        
        client_id = self._get_client_id(request)
        
        # Check rate limit
        allowed = await self._check_rate_limit(client_id)
        if not allowed:
            LOGGER.warning(
                "Rate limit exceeded",
                extra={"context": {"client_id": client_id}},
            )
            raise HTTPException(
                status_code=429,
                detail="Rate limit exceeded. Try again later.",
                headers={"Retry-After": "60"},
            )
        
        # Process request
        response = await call_next(request)
        
        # Add rate limit headers
        response.headers["X-RateLimit-Limit"] = str(self.config.requests_per_minute)
        remaining = await self._get_remaining(client_id)
        response.headers["X-RateLimit-Remaining"] = str(remaining)
        
        return response
    
    def _get_client_id(self, request: Request) -> str:
        """Extract client identifier for rate limiting.
        
        Args:
            request: HTTP request
            
        Returns:
            Client identifier (user ID or IP address)
        """
        # Check for authenticated user context
        if hasattr(request.state, "user"):
            return f"user:{request.state.user.id}"
        
        # Fall back to IP address
        client_host = request.client.host if request.client else "unknown"
        return f"ip:{client_host}"
    
    async def _check_rate_limit(self, client_id: str) -> bool:
        """Check if request is allowed under rate limit.
        
        Args:
            client_id: Client identifier
            
        Returns:
            True if request is allowed, False otherwise
        """
        if self.backend == "redis":
            return await self._check_rate_limit_redis(client_id)
        return await self._check_rate_limit_memory(client_id)
    
    async def _check_rate_limit_memory(self, client_id: str) -> bool:
        """Memory-backed token bucket rate limiting.
        
        Args:
            client_id: Client identifier
            
        Returns:
            True if request is allowed, False otherwise
        """
        bucket = self.buckets[client_id]
        now = time.time()
        
        # Refill tokens based on time elapsed
        elapsed = now - bucket["last_update"]
        refill = (elapsed * self.config.requests_per_minute) / 60.0
        bucket["tokens"] = min(self.config.burst_size, bucket["tokens"] + refill)
        bucket["last_update"] = now
        
        # Check if request can proceed
        if bucket["tokens"] >= 1:
            bucket["tokens"] -= 1
            return True
        return False
    
    async def _check_rate_limit_redis(self, client_id: str) -> bool:
        """Redis-backed rate limiting using sliding window.
        
        Args:
            client_id: Client identifier
            
        Returns:
            True if request is allowed, False otherwise
        """
        key = f"ratelimit:{client_id}"
        now = int(time.time())
        window = 60  # seconds
        
        pipe = self.redis.pipeline()
        pipe.zadd(key, {str(now): now})
        pipe.zremrangebyscore(key, 0, now - window)
        pipe.zcard(key)
        pipe.expire(key, window + 1)
        results = await pipe.execute()
        
        request_count = results[2]
        return request_count <= self.config.requests_per_minute
    
    async def _get_remaining(self, client_id: str) -> int:
        """Get remaining requests for client.
        
        Args:
            client_id: Client identifier
            
        Returns:
            Number of remaining requests in current window
        """
        if self.backend == "memory":
            bucket = self.buckets.get(client_id)
            if bucket:
                return int(bucket["tokens"])
            return self.config.burst_size
        
        # For Redis, calculate from sliding window
        key = f"ratelimit:{client_id}"
        now = int(time.time())
        window = 60
        
        count = await self.redis.zcount(key, now - window, now)
        return max(0, self.config.requests_per_minute - count)



"""Graph relationship builders.

These helpers infer edges between already-enumerated Microsoft Graph nodes.
"""

from __future__ import annotations

from typing import Dict, Iterable, List

from ..adt_types import ResourceNode, ResourceRelationship


def build_graph_appid_edges(nodes: Iterable[ResourceNode]) -> List[ResourceRelationship]:
    """Infer Application <-> ServicePrincipal relationships via appId.

    Emits directed edges from servicePrincipal -> application when both share the
    same appId.
    """

    apps_by_appid: Dict[str, ResourceNode] = {}
    sps_by_appid: Dict[str, ResourceNode] = {}

    for n in nodes:
        app_id = n.tags.get("appId") if n.tags else None
        if not app_id:
            continue

        if n.type == "Microsoft.Graph/Application":
            apps_by_appid[str(app_id)] = n
        elif n.type == "Microsoft.Graph/ServicePrincipal":
            sps_by_appid[str(app_id)] = n

    edges: List[ResourceRelationship] = []
    for app_id, sp in sorted(sps_by_appid.items(), key=lambda kv: kv[0]):
        app = apps_by_appid.get(app_id)
        if not app:
            continue
        edges.append(
            ResourceRelationship(
                source_id=sp.id,
                target_id=app.id,
                relation_type="appId",
                weight=1.0,
            )
        )

    return edges
